import heapq
from typing import Optional, Dict, Any, Tuple

from ..base import BaseSimObj


class EventQueue(BaseSimObj):
    """ Queue which stores simulation events.

    Args:
        events (List[Event]): A list of Event-like objects.
    """

    def __init__(self, events=None):
        self._queue = []
        if events is not None:
            self.add_events(events)
        self._timestep = 0

    def __len__(self):
        """ Return the length of this queue.

        Returns:
            int: Length of the queue.
        """
        return len(self._queue)

    @property
    def queue(self):
        """ Return the queue of events """
        return self._queue

    def empty(self):
        """ Return if the queue is empty.

        Returns:
            bool: True if the queue is empty.
        """
        return len(self._queue) == 0

    def add_event(self, event):
        """ Add an event to the queue.

        Args:
            event (Event like): An Event-like object.

        Returns:
            None
        """
        heapq.heappush(self._queue, (event.timestamp, event))

    def add_events(self, events):
        """ Add multiple events at a time to the queue.

        Args:
            events (List[Event like]): A list of Event-like objects.

        Returns:
            None
        """
        for e in events:
            self.add_event(e)

    def get_event(self):
        """ Return the next event in the queue.

        Returns:
            Event like: The next event in the queue.
        """
        return heapq.heappop(self._queue)[1]

    def get_current_events(self, timestep):
        """ Return all events occurring before or during timestep.

        Args:
            timestep (int): Time index in periods.

        Returns:
            List[Event]: List of all events occurring before or during timestep.
        """
        self._timestep = timestep
        current_events = []
        while not self.empty() and self._queue[0][0] <= self._timestep:
            current_events.append(self.get_event())
        return current_events

    def get_last_timestamp(self):
        """ Return the timestamp of the last event (chronologically) in the event queue

        Returns:
            int: Last timestamp in the event queue, or None if the
                event queue is empty.
        """
        if not self.empty():
            return max(self._queue, key=lambda x: x[0])[0]
        else:
            return None

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict = {"_timestep": self._timestep}

        event_queue = []
        for (ts, event) in self._queue:
            # noinspection PyProtectedMember
            registry, context_dict = event._to_registry(context_dict=context_dict)
            event_queue.append((ts, registry["id"]))
        attribute_dict["_queue"] = event_queue

        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls()
        out_obj._timestep = attribute_dict["_timestep"]

        event_queue = []
        for (ts, event) in attribute_dict["_queue"]:
            # noinspection PyProtectedMember
            loaded_event, loaded_dict = BaseSimObj._build_from_id(
                event, context_dict, loaded_dict=loaded_dict
            )
            event_queue.append((ts, loaded_event))
        out_obj._queue = event_queue

        return out_obj, loaded_dict



# coding=utf-8
""" Tests for stochastic event generation. """
import unittest
from unittest.mock import Mock, call, patch
from acnportal.acnsim.events.stochastic_events import StochasticEvents
import numpy as np
from datetime import datetime
from acnportal.acnsim import Battery, EV
from typing import List


class TestStochasticEvents(unittest.TestCase):
    def setUp(self) -> None:
        """ Tests for StochasticEvents. """
        self.period = 5
        self.periods_per_hour = 60 / self.period
        self.periods_per_day = self.periods_per_hour * 24
        self.voltage = 208
        self.max_battery_power = 7
        self.sessions_per_day = [3]

    def _default_tests(self, evs: List[EV]):
        with self.subTest("test_length_of_evs"):
            self.assertEqual(3, len(evs))

        with self.subTest("test_arrival_times"):
            self.assertListEqual([ev.arrival for ev in evs], [78, 99, 120])

        with self.subTest("test_station_ids"):
            self.assertListEqual(
                [ev.station_id for ev in evs], [f"station_{i}" for i in range(3)]
            )

        with self.subTest("test_session_ids"):
            self.assertListEqual(
                [ev.session_id for ev in evs], [f"session_{i}" for i in range(3)]
            )

        with self.subTest("test_battery"):
            for ev in evs:
                self.assertIsInstance(ev._battery, Battery)
                self.assertEqual(ev.maximum_charging_power, self.max_battery_power)

    @patch.object(StochasticEvents, "sample")
    def test_generate_events(self, sample_mock: Mock) -> None:
        samples = np.array([[6.5, 8, 10], [8.3, 6.05, 3], [10, 3, 6.6]])
        gen = StochasticEvents()
        sample_mock.return_value = samples
        queue = gen.generate_events(
            self.sessions_per_day, self.period, self.voltage, self.max_battery_power
        )
        evs = [event[1].ev for event in queue.queue]
        self._default_tests(evs)

        with self.subTest("test_departure_time"):
            self.assertListEqual([ev.departure for ev in evs], [174, 172, 156])

        with self.subTest("test_requested_energy"):
            self.assertListEqual([ev.requested_energy for ev in evs], [10, 3, 6.6])

    @patch.object(StochasticEvents, "sample")
    def test_generate_events_with_force_feasible(self, sample_mock: Mock) -> None:
        samples = np.array([[6.5, 1, 10], [8.3, 6.05, 3], [10, 3, 6.6]])
        gen = StochasticEvents()
        sample_mock.return_value = samples
        queue = gen.generate_events(
            self.sessions_per_day,
            self.period,
            self.voltage,
            self.max_battery_power,
            force_feasible=True,
        )
        evs = [event[1].ev for event in queue.queue]
        self._default_tests(evs)

        with self.subTest("test_departure_time"):
            self.assertListEqual([ev.departure for ev in evs], [90, 172, 156])

        with self.subTest("test_requested_energy"):
            self.assertListEqual(
                [ev.requested_energy for ev in evs], [self.max_battery_power, 3, 6.6]
            )

    @patch.object(StochasticEvents, "sample")
    def test_generate_events_with_max_len(self, sample_mock: Mock) -> None:
        samples = np.array([[6.5, 8, 10], [8.3, 6.05, 3], [10, 3, 6.6]])
        gen = StochasticEvents()
        sample_mock.return_value = samples
        queue = gen.generate_events(
            self.sessions_per_day,
            self.period,
            self.voltage,
            self.max_battery_power,
            max_len=3,
        )
        evs = [event[1].ev for event in queue.queue]
        self._default_tests(evs)

        with self.subTest("test_departure_time"):
            self.assertListEqual([ev.departure for ev in evs], [114, 135, 156])

        with self.subTest("test_requested_energy"):
            self.assertListEqual([ev.requested_energy for ev in evs], [10, 3, 6.6])

    @patch.object(StochasticEvents, "sample")
    def test_generate_events_with_max_len_and_force_feasible(
        self, sample_mock: Mock
    ) -> None:
        samples = np.array([[6.5, 8, 10], [8.3, 6.05, 3], [10, 3, 15]])
        gen = StochasticEvents()
        sample_mock.return_value = samples
        queue = gen.generate_events(
            self.sessions_per_day,
            self.period,
            self.voltage,
            self.max_battery_power,
            max_len=1,
            force_feasible=True,
        )
        evs = [event[1].ev for event in queue.queue]
        self._default_tests(evs)

        with self.subTest("test_departure_time"):
            self.assertListEqual([ev.departure for ev in evs], [90, 111, 132])

        with self.subTest("test_requested_energy"):
            self.assertListEqual(
                [ev.requested_energy for ev in evs],
                [self.max_battery_power, 3, self.max_battery_power],
            )

    @patch.object(StochasticEvents, "sample")
    def test_generate_events_multi_day(self, sample_mock: Mock) -> None:
        samples = np.array([[6.5, 8, 10], [8.3, 6.05, 3], [10, 3, 15]])
        gen = StochasticEvents()

        # sample expects an argument n_samples. We need to copy the array to prevent
        # changing it in the function between calls to sample.
        sample_mock.side_effect = lambda n: samples.copy()
        queue = gen.generate_events(
            [3, 3], self.period, self.voltage, self.max_battery_power
        )

        sample_mock.assert_has_calls([call(3), call(3)])

        evs = [event[1].ev for event in queue.queue]
        with self.subTest("test_length_of_evs"):
            self.assertEqual(6, len(evs))

        with self.subTest("test_arrival_times"):
            self.assertListEqual(
                [ev.arrival for ev in evs],
                [
                    78,
                    99,
                    120,
                    self.periods_per_day + 78,
                    self.periods_per_day + 99,
                    self.periods_per_day + 120,
                ],
            )

        with self.subTest("test_departure_time"):
            self.assertListEqual(
                [ev.departure for ev in evs],
                [
                    174,
                    172,
                    156,
                    self.periods_per_day + 174,
                    self.periods_per_day + 172,
                    self.periods_per_day + 156,
                ],
            )

    @patch.object(StochasticEvents, "sample")
    def test_generate_events_multi_day_with_0_sessions(self, sample_mock: Mock) -> None:
        samples = np.array([[6.5, 8, 10], [8.3, 6.05, 3], [10, 3, 15]])
        gen = StochasticEvents()

        # sample expects an argument n_samples. We need to copy the array to prevent
        # changing it in the function between calls to sample.
        sample_mock.side_effect = lambda n: samples.copy()
        queue = gen.generate_events(
            [3, 0, 3], self.period, self.voltage, self.max_battery_power
        )

        sample_mock.assert_has_calls([call(3), call(3)])

        evs = [event[1].ev for event in queue.queue]
        with self.subTest("test_length_of_evs"):
            self.assertEqual(6, len(evs))

        with self.subTest("test_arrival_times"):
            self.assertListEqual(
                [ev.arrival for ev in evs],
                [
                    78,
                    99,
                    120,
                    2 * self.periods_per_day + 78,
                    2 * self.periods_per_day + 99,
                    2 * self.periods_per_day + 120,
                ],
            )

        with self.subTest("test_departure_time"):
            self.assertListEqual(
                [ev.departure for ev in evs],
                [
                    174,
                    172,
                    156,
                    2 * self.periods_per_day + 174,
                    2 * self.periods_per_day + 172,
                    2 * self.periods_per_day + 156,
                ],
            )

    def test_extract_training_data(self) -> None:
        sessions = [
            {
                "connectionTime": datetime(2020, 1, 1),
                "disconnectTime": datetime(2020, 1, 1, 8, 30),
                "kWhDelivered": 8.24,
            },
            {
                "connectionTime": datetime(2020, 1, 1, 7, 24),
                "disconnectTime": datetime(2020, 1, 1, 10),
                "kWhDelivered": 1,
            },
            {
                "connectionTime": datetime(2020, 1, 1, 8),
                "disconnectTime": datetime(2020, 1, 1, 13),
                "kWhDelivered": 12.3,
            },
        ]
        expected = np.array([[0, 8.5, 8.24], [7.4, 2.6, 1], [8, 5, 12.3]])
        np.testing.assert_equal(
            StochasticEvents.extract_training_data(sessions), expected
        )

    def test_clip_samples(self) -> None:
        samples = np.array(
            [
                [8, 8.5, 8.24],
                [-1, 8.5, 8.24],
                [26, 8.5, 8.24],
                [8, -1, 8.24],
                [8, 72, 8.24],
                [8, 8.5, -11],
                [8, 8.5, 200],
            ]
        )

        expected = np.array(
            [
                [8, 8.5, 8.24],
                [0, 8.5, 8.24],
                [24, 8.5, 8.24],
                [8, 0.0833, 8.24],
                [8, 48, 8.24],
                [8, 8.5, 0.5],
                [8, 8.5, 150],
            ]
        )

        se = StochasticEvents()
        np.testing.assert_equal(se.clip_samples(samples), expected)


if __name__ == "__main__":
    unittest.main()



import json
import unittest

import numpy as np

from acnportal.acnsim import ChargingNetwork
from acnportal.acnsim.base import ErrorAllWrapper, NpEncoder


class TestErrorAllWrapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.builtin_data = 2
        cls.builtin_wrapper = ErrorAllWrapper(cls.builtin_data)
        cls.network_data = ChargingNetwork()
        cls.network_wrapper = ErrorAllWrapper(cls.network_data)

    def test_correct_on_init(self):
        self.assertEqual(self.builtin_data, self.builtin_wrapper.data)
        self.assertEqual(self.network_data, self.network_wrapper.data)

    def test_builtin_success(self):
        self.assertEqual(self.builtin_wrapper.__dict__, {"_data": 2})

    def test_builtin_error(self):
        with self.assertRaises(TypeError):
            _ = self.builtin_wrapper == self.network_wrapper

    def test_network_attribute_error(self):
        with self.assertRaises(TypeError):
            _ = self.network_wrapper.constraint_matrix

    def test_network_function_error(self):
        with self.assertRaises(TypeError):
            _ = self.network_wrapper.is_feasible(np.array([[0]]))


class TestNpEncoder(unittest.TestCase):
    def test_integer_dump(self):
        test_uint32 = np.uint32(2)
        self.assertEqual(json.dumps(test_uint32, cls=NpEncoder), "2")

    def test_floating_dump(self):
        test_float32 = np.float32(2)
        self.assertEqual(json.dumps(test_float32, cls=NpEncoder), "2.0")

    def test_ndarray_dump(self):
        test_ndarray = np.array([1, 2])
        self.assertEqual(json.dumps(test_ndarray, cls=NpEncoder), "[1, 2]")


if __name__ == "__main__":
    unittest.main()



from .. import ChargingNetwork
from ..current import Current
from ...models.evse import get_evse_by_type


def office001_acn(
    basic_evse=False, voltage=208, transformer_cap=50, network_type=ChargingNetwork
):
    """ Predefined ChargingNetwork for the Office001 ACN.

    Args:
        basic_evse (bool): If True use BASIC EVSE type instead of actual AeroViroment types.
        voltage (float): Default voltage at the EVSEs. Does not affect the current rating of the transformer which is
            based on nominal voltages in the network. 277V LL for delta primary and 120V LN for wye secondary. [V]
        transformer_cap (float): Capacity of the transformer feeding the network. Default: 150. [kW]
        network_type (ChargingNetwork like): Type to be returned. Should be ChargingNetwork or a subclass.

    Returns:
        ChargingNetwork: A ChargingNetwork-like object configured with the EVSEs and constraints of the Office 1 ACN.
    """
    network = network_type()

    if basic_evse:
        evse_type = "BASIC"
    else:
        evse_type = "AeroVironment"

    # Define the sets of EVSEs
    AB_ids = ["01", "04", "07"]
    BC_ids = ["02", "05", "08"]
    CA_ids = ["03", "06"]

    # Add EVSEs
    for evse_id in AB_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type), voltage, 30)
    for evse_id in BC_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type), voltage, -90)
    for evse_id in CA_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type), voltage, 150)

    # Define currents
    AB = Current(AB_ids)
    BC = Current(BC_ids)
    CA = Current(CA_ids)

    # Define intermediate currents
    I3a = AB - CA
    I3b = BC - AB
    I3c = CA - BC
    I2a = (1 / 4) * (I3a - I3c)
    I2b = (1 / 4) * (I3b - I3a)
    I2c = (1 / 4) * (I3c - I3b)

    # Build constraint set
    primary_side_constr = transformer_cap * 1000 / 3 / 277
    secondary_side_constr = transformer_cap * 1000 / 3 / 120
    network.add_constraint(I3a, secondary_side_constr, name="Secondary A")
    network.add_constraint(I3b, secondary_side_constr, name="Secondary B")
    network.add_constraint(I3c, secondary_side_constr, name="Secondary C")
    network.add_constraint(I2a, primary_side_constr, name="Primary A")
    network.add_constraint(I2b, primary_side_constr, name="Primary B")
    network.add_constraint(I2c, primary_side_constr, name="Primary C")

    return network



from unittest import TestCase

import pandas as pd

from acnportal.acnsim import Current


class TestCurrent(TestCase):
    def setUp(self):
        pass

    def test_init_loads_dict_input(self):
        curr_dict = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        self.current = Current(curr_dict)
        pd.testing.assert_series_equal(
            self.current,
            pd.Series([0.25, 0.50, -0.25], index=["PS-001", "PS-002", "PS-003"]),
            check_series_type=False,
        )

    def test_init_str_load_input(self):
        curr_str = "PS-001"
        self.current = Current(curr_str)
        pd.testing.assert_series_equal(
            self.current, pd.Series([1], index=["PS-001"]), check_series_type=False
        )

    def test_init_none_input(self):
        self.current = Current()
        # The object type is specified explicitly here to address a
        # warning in pandas 1.x.
        pd.testing.assert_series_equal(
            self.current, pd.Series(dtype="float64"), check_series_type=False
        )

    def test_init_str_lst_input(self):
        curr_strs = ["PS-001", "PS-002", "PS-003"]
        self.current = Current(curr_strs)
        pd.testing.assert_series_equal(
            self.current,
            pd.Series([1, 1, 1], index=["PS-001", "PS-002", "PS-003"]),
            check_series_type=False,
        )

    def test_add_current_equal_station_ids(self):
        curr_dict1 = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        self.current1 = Current(curr_dict1)
        curr_dict2 = {"PS-001": 0.30, "PS-002": -0.60, "PS-003": 0.50}
        self.current2 = Current(curr_dict2)
        self.sum_curr = self.current1 + self.current2
        self.assertIsInstance(self.sum_curr, Current)
        pd.testing.assert_series_equal(
            self.sum_curr,
            pd.Series([0.55, -0.10, 0.25], index=["PS-001", "PS-002", "PS-003"]),
            check_series_type=False,
        )

    def test_add_current_unequal_station_ids(self):
        curr_dict1 = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        self.current1 = Current(curr_dict1)
        curr_dict2 = {"PS-006": 0.30, "PS-004": -0.60, "PS-002": 0.50}
        self.current2 = Current(curr_dict2)
        self.sum_curr = self.current1 + self.current2
        self.assertIsInstance(self.sum_curr, Current)
        pd.testing.assert_series_equal(
            self.sum_curr,
            pd.Series(
                [0.25, 1.00, -0.25, -0.60, 0.30],
                index=["PS-001", "PS-002", "PS-003", "PS-004", "PS-006"],
            ),
            check_series_type=False,
        )

    def test_sub_current_unequal_station_ids(self):
        curr_dict1 = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        self.current1 = Current(curr_dict1)
        curr_dict2 = {"PS-006": 0.30, "PS-004": -0.60, "PS-002": 0.50}
        self.current2 = Current(curr_dict2)
        self.diff_curr = self.current1 - self.current2
        self.assertIsInstance(self.diff_curr, Current)
        pd.testing.assert_series_equal(
            self.diff_curr,
            pd.Series(
                [0.25, 0.00, -0.25, 0.60, -0.30],
                index=["PS-001", "PS-002", "PS-003", "PS-004", "PS-006"],
            ),
            check_series_type=False,
        )

    def test_mul_current(self):
        curr_dict = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        self.current = Current(curr_dict)
        self.current *= 2
        self.assertIsInstance(self.current, Current)
        pd.testing.assert_series_equal(
            self.current,
            pd.Series([0.50, 1.00, -0.5], index=["PS-001", "PS-002", "PS-003"]),
            check_series_type=False,
        )



# coding=utf-8
""" Functions to generate Sessions and Infrastructures for testing algorithms. """
from typing import List, Iterable, Optional, Dict, Union

import numpy as np

SessionDict = Dict[str, Optional[Union[str, float, int, Iterable[float]]]]


def session_generator(
    num_sessions: int,
    arrivals: List[int],
    departures: List[int],
    requested_energy: List[float],
    remaining_energy: List[float],
    max_rates: List[Union[float, Iterable[float], np.ndarray]],
    min_rates: Optional[List[Union[float, Iterable[float], np.ndarray]]] = None,
    station_ids: Optional[List[str]] = None,
    estimated_departures: Optional[List[Union[float, Iterable[float], np.ndarray]]] = None
) -> List[SessionDict]:
    """ Generate Sessions with the input info in dict format. """
    sessions: List[SessionDict] = []
    min_rates: List[float] = min_rates if min_rates is not None else [0] * num_sessions
    for i in range(num_sessions):
        station_id: str = station_ids[i] if station_ids is not None else f"{i}"
        session_id: str = f"{i}"
        s: SessionDict = {
            "station_id": station_id,
            "session_id": session_id,
            "requested_energy": requested_energy[i],
            "energy_delivered": requested_energy[i] - remaining_energy[i],
            "arrival": arrivals[i],
            "departure": departures[i],
            "estimated_departure": estimated_departures[i] if estimated_departures is not None else departures[i],
            "min_rates": min_rates[i],
            "max_rates": max_rates[i],
        }
        sessions.append(s)
    return sessions


InfrastructureDict = Dict[str, Union[np.ndarray, List[str], List[np.ndarray]]]


def single_phase_single_constraint(
    num_evses: int,
    limit: float,
    max_pilot: float = 32,
    min_pilot: float = 8,
    allowable_pilots: Optional[List[np.ndarray]] = None,
    is_continuous: Optional[np.ndarray] = None,
) -> InfrastructureDict:
    """ Generates a single-phase one-constraint network; returns the corresponding
    infrastructure info in a dict. """
    if allowable_pilots is None:
        allowable_pilots: List[np.ndarray] = [
            np.array([min_pilot, max_pilot])
        ] * num_evses
    if is_continuous is None:
        is_continuous: np.ndarray = np.ones(num_evses, dtype=bool)
    infrastructure: InfrastructureDict = {
        "constraint_matrix": np.ones((1, num_evses)),
        "constraint_limits": np.array([limit]),
        "phases": np.zeros(num_evses),
        "voltages": np.repeat(208, num_evses),
        "constraint_ids": ["all"],
        "station_ids": [f"{i}" for i in range(num_evses)],
        "max_pilot": np.repeat(max_pilot, num_evses),
        "min_pilot": np.repeat(min_pilot, num_evses),
        "allowable_pilots": allowable_pilots,
        "is_continuous": is_continuous,
    }
    return infrastructure


def three_phase_balanced_network(
    evses_per_phase: int,
    limit: float,
    max_pilot: float = 32,
    min_pilot: float = 8,
    allowable_pilots: Optional[List[Iterable[float]]] = None,
    is_continuous: Optional[np.ndarray] = None,
) -> InfrastructureDict:
    """ Generates a 3-phase 3-constraint network; returns the corresponding
    infrastructure info in a dict. """
    n: int = evses_per_phase
    num_evses: int = 3 * evses_per_phase
    if allowable_pilots is None:
        allowable_pilots: List[np.ndarray] = [
            np.array([min_pilot, max_pilot])
        ] * num_evses
    if is_continuous is None:
        is_continuous: np.ndarray = np.ones(num_evses, dtype=bool)
    infrastructure: InfrastructureDict = {
        "constraint_matrix": np.array(
            [
                [1] * n + [-1] * n + [0] * n,
                [0] * n + [1] * n + [-1] * n,
                [-1] * n + [0] * n + [1] * n,
            ]
        ),
        "constraint_limits": np.repeat(limit, 3),
        "phases": np.array([0] * n + [-120] * n + [120] * n),
        "voltages": np.repeat(208, 3 * evses_per_phase),
        "constraint_ids": ["AB", "BC", "CA"],
        "station_ids": [f"{i}" for i in range(num_evses)],
        "max_pilot": np.repeat(max_pilot, num_evses),
        "min_pilot": np.repeat(min_pilot, num_evses),
        "allowable_pilots": allowable_pilots,
        "is_continuous": is_continuous,
    }
    return infrastructure



from .. import ChargingNetwork
from ..current import Current
from ...models.evse import get_evse_by_type


def caltech_acn(
    basic_evse=False, voltage=208, transformer_cap=150, network_type=ChargingNetwork
):
    """ Predefined ChargingNetwork for the Caltech ACN.

    Args:
        basic_evse (bool): If True use BASIC EVSE type instead of actual AeroViroment and ClipperCreek types.
        voltage (float): Default voltage at the EVSEs. Does not affect the current rating of the transformer which is
            based on nominal voltages in the network. 277V LL for delta primary and 120V LN for wye secondary. [V]
        transformer_cap (float): Capacity of the transformer in the CaltechACN. Default: 150. [kW]
        network_type (ChargingNetwork like): Type to be returned. Should be ChargingNetwork or a subclass.

    Returns:
        ChargingNetwork: A ChargingNetwork-like object configured with the EVSEs and constraints of the Caltech ACN.
    """
    network = network_type()

    if basic_evse:
        evse_type = {"AV": "BASIC", "CC": "BASIC"}
    else:
        evse_type = {"AV": "AeroVironment", "CC": "ClipperCreek"}

    # Define the sets of EVSEs in the Caltech ACN.
    CC_pod_ids = [
        "CA-322",
        "CA-493",
        "CA-496",
        "CA-320",
        "CA-495",
        "CA-321",
        "CA-323",
        "CA-494",
    ]
    AV_pod_ids = [
        "CA-324",
        "CA-325",
        "CA-326",
        "CA-327",
        "CA-489",
        "CA-490",
        "CA-491",
        "CA-492",
    ]
    AB_ids = (
        ["CA-{0}".format(i) for i in [308, 508, 303, 513, 310, 506, 316, 500, 318, 498]]
        + AV_pod_ids
        + CC_pod_ids
    )
    BC_ids = [
        "CA-{0}".format(i)
        for i in [304, 512, 305, 511, 313, 503, 311, 505, 317, 499, 148, 149, 212, 213]
    ]
    CA_ids = [
        "CA-{0}".format(i)
        for i in [307, 509, 309, 507, 306, 510, 315, 501, 319, 497, 312, 504, 314, 502]
    ]

    # Add Caltech EVSEs
    for evse_id in AB_ids:
        if evse_id not in CC_pod_ids:
            network.register_evse(
                get_evse_by_type(evse_id, evse_type["AV"]), voltage, 30
            )
        else:
            network.register_evse(
                get_evse_by_type(evse_id, evse_type["CC"]), voltage, 30
            )
    for evse_id in BC_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type["AV"]), voltage, -90)
    for evse_id in CA_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type["AV"]), voltage, 150)

    # Add Caltech Constraint Set
    CC_pod = Current(CC_pod_ids)
    AV_pod = Current(AV_pod_ids)
    AB = Current(AB_ids)
    BC = Current(BC_ids)
    CA = Current(CA_ids)

    # Define intermediate currents
    I3a = AB - CA
    I3b = BC - AB
    I3c = CA - BC
    I2a = (1 / 4) * (I3a - I3c)
    I2b = (1 / 4) * (I3b - I3a)
    I2c = (1 / 4) * (I3c - I3b)

    # Build constraint set
    primary_side_constr = transformer_cap * 1000 / 3 / 277
    secondary_side_constr = transformer_cap * 1000 / 3 / 120
    network.add_constraint(CC_pod, 80, name="CC Pod")
    network.add_constraint(AV_pod, 80, name="AV Pod")
    network.add_constraint(I3a, secondary_side_constr, name="Secondary A")
    network.add_constraint(I3b, secondary_side_constr, name="Secondary B")
    network.add_constraint(I3c, secondary_side_constr, name="Secondary C")
    network.add_constraint(I2a, primary_side_constr, name="Primary A")
    network.add_constraint(I2b, primary_side_constr, name="Primary B")
    network.add_constraint(I2c, primary_side_constr, name="Primary C")

    return network


def CaltechACN(
    basic_evse=False, voltage=208, transformer_cap=150, network_type=ChargingNetwork
):
    """ Wrapper around caltech_acn for backward compatibility. Will be removed in future release. """
    print(
        "WARNING: CaltechACN will be removed in a future release. Please use caltech_acn()."
    )
    return caltech_acn(basic_evse, voltage, transformer_cap, network_type)



from .event import Event
from .event import PluginEvent
from .event import UnplugEvent
from .event import RecomputeEvent
from .event_queue import EventQueue
from .acndata_events import *
from .stochastic_events import *



from unittest import TestCase

from acnportal.acnsim.events import EventQueue, Event


class TestEventAccessors(TestCase):
    def setUp(self):
        self.event = Event(5)
        self.assertEqual(self.event.event_type, "")

    def test_type_deprecation_warning(self):
        self.event = Event(5)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(self.event.type, "")


class TestEventQueue(TestCase):
    def setUp(self):
        self.events = EventQueue()

    def test_empty_on_init(self):
        self.assertTrue(self.events.empty())

    def test_add_event(self):
        self.events.add_event(Event(5))
        self.assertFalse(self.events.empty())
        self.assertEqual(len(self.events._queue), 1)

    def test_add_events(self):
        events = [Event(i) for i in range(1, 6)]
        self.events.add_events(events)
        self.assertFalse(self.events.empty())
        self.assertEqual(5, len(self.events._queue))

    def test_len(self):
        events = [Event(i) for i in range(1, 6)]
        self.events.add_events(events)
        self.assertEqual(5, len(self.events))

    def test_get_event(self):
        self.events.add_event(Event(5))
        self.assertEqual(len(self.events._queue), 1)
        e = self.events.get_event()
        self.assertTrue(self.events.empty())
        self.assertEqual(5, e.timestamp)

    def test_get_current_events_empty(self):
        events = []
        self.events.add_events(events)
        curr_events = self.events.get_current_events(3)
        self.assertEqual(curr_events, [])

    def test_get_current_events(self):
        events = [Event(i) for i in range(1, 6)]
        self.events.add_events(events)
        curr_events = self.events.get_current_events(3)
        self.assertEqual(len(curr_events), 3)
        for event, timestamp in zip(curr_events, [1, 2, 3]):
            self.assertEqual(event.timestamp, timestamp)

    def test_get_last_timestamp(self):
        events = [Event(i) for i in range(1, 6)]
        self.events.add_events(events)
        self.assertEqual(5, self.events.get_last_timestamp())
        self.events.add_event(Event(8))
        self.assertEqual(8, self.events.get_last_timestamp())
        # Check that the last timestamp is unaltered from a call to
        # get_current_events at an earlier timestamp.
        _ = self.events.get_current_events(3)
        self.assertEqual(8, self.events.get_last_timestamp())

    def test_get_last_timestamp_no_events(self):
        events = []
        self.events.add_events(events)
        self.assertIsNone(self.events.get_last_timestamp())



# coding=utf-8
""" Class to represent the physical model of an adaptive charging network. """
from typing import Optional, List, Dict, Any, Tuple

from .current import Current
import pandas as pd
import numpy as np
from collections import OrderedDict
import warnings

from ..models import EV, BaseEVSE
from ..base import BaseSimObj


class ChargingNetwork(BaseSimObj):
    """
    The ChargingNetwork class describes the infrastructure of the
    charging network with information about the types of the charging
    station_schedule.

    Args:
        violation_tolerance (float): Absolute amount by which an input
            charging schedule may violate network constraints (A).
        relative_tolerance (float): Relative amount by which an input
            charging schedule may violate network constraints (A).
    """

    _EVSEs: OrderedDict
    constraint_matrix: Optional[np.ndarray]
    magnitudes: np.ndarray
    constraint_index: List[str]
    _voltages: np.ndarray
    _phase_angles: np.ndarray
    violation_tolerance: float
    relative_tolerance: float
    _station_ids_dict: Dict[str, int]
    max_pilot_signals: np.ndarray
    min_pilot_signals: np.ndarray
    allowable_rates: List[np.ndarray]
    is_continuous: np.ndarray

    def __init__(
        self, violation_tolerance: float = 1e-5, relative_tolerance: float = 1e-7
    ):
        self._EVSEs = OrderedDict()
        # Matrix of constraints
        self.constraint_matrix = None
        # Vector of limiting magnitudes
        self.magnitudes = np.array([])
        # List of constraints in order of addition to network
        self.constraint_index = []
        self._voltages = np.array([])
        self._phase_angles = np.array([])
        self.violation_tolerance = violation_tolerance
        self.relative_tolerance = relative_tolerance

        # Cached information-storing objects for use by Interface.
        (
            self._station_ids_dict,
            self.max_pilot_signals,
            self.min_pilot_signals,
            self.allowable_rates,
            self.is_continuous,
        ) = self._update_info_store()

    def _update_info_store(
        self,
    ) -> Tuple[Dict[str, int], np.ndarray, np.ndarray, List[np.ndarray], np.ndarray]:
        station_ids: List[str] = self.station_ids
        self._station_ids_dict = {
            station_id: i for i, station_id in enumerate(station_ids)
        }
        self.max_pilot_signals = np.array(
            [self._EVSEs[station_id].max_rate for station_id in station_ids]
        )
        self.min_pilot_signals = np.array(
            [self._EVSEs[station_id].min_rate for station_id in station_ids]
        )
        allowable_rates = []
        is_continuous = []
        for station_id in station_ids:
            # Get allowable pilot signals and continuity for this EVSE.
            evse = self._EVSEs[station_id]
            continuous, allowable = evse.is_continuous, evse.allowable_pilot_signals
            allowable_rates.append(np.array(allowable))
            is_continuous.append(continuous)
        is_continuous = np.array(is_continuous, dtype="bool")
        self.allowable_rates = allowable_rates
        self.is_continuous = is_continuous
        return (
            self._station_ids_dict,
            self.max_pilot_signals,
            self.min_pilot_signals,
            self.allowable_rates,
            self.is_continuous,
        )

    @property
    def current_charging_rates(self) -> np.ndarray:
        """ Return the current actual charging rate of all EVSEs in the network. If
        no EV is attached to a given EVSE, that EVSE's charging rate is 0. In the
        returned array, the charging rates are given in the same order as the list of
        EVSEs given by station_ids

        Returns:
            np.Array: numpy ndarray of actual charging rates of all EVSEs in the
                network.
        """
        return np.array(
            [
                evse.ev.current_charging_rate if evse.ev is not None else 0
                for evse in self._EVSEs.values()
            ]
        )

    @property
    def station_ids(self) -> List[str]:
        """ Return the IDs of all registered EVSEs.

        Returns:
            List[str]: List of all registered EVSE IDs.
        """
        return list(self._EVSEs.keys())

    @property
    def active_evs(self) -> List[EV]:
        """ Return all EVs which are connected to an EVSE and which are not already
        fully charged.

        Returns:
            List[EV]: List of EVs which can currently be charged.
        """
        return [
            evse.ev
            for evse in self._EVSEs.values()
            if evse.ev is not None and not evse.ev.fully_charged
        ]

    @property
    def active_station_ids(self) -> List[str]:
        """ Return IDs for all stations which have an active EV attached.

        Returns:
            List[str]: List of the station_id of all stations which have an
                active EV attached.
        """
        return [
            evse.station_id
            for evse in self._EVSEs.values()
            if evse.ev is not None and not evse.ev.fully_charged
        ]

    @property
    def voltages(self) -> Dict[str, float]:
        """ Return dictionary of voltages for all EVSEs in the network.

        Returns:
            Dict[str, float]: Dictionary mapping EVSE ids their input voltage. [V]
        """
        return {
            self.station_ids[i]: self._voltages[i] for i in range(len(self._voltages))
        }

    @property
    def phase_angles(self) -> Dict[str, float]:
        """ Return dictionary of phase angles for all EVSEs in the network.

        Returns:
            Dict[str, float]: Dictionary mapping EVSE ids their input phase angle. [
                degrees]
        """
        return {
            self.station_ids[i]: self._phase_angles[i]
            for i in range(len(self._phase_angles))
        }

    def register_evse(self, evse: BaseEVSE, voltage: float, phase_angle: float) -> None:
        """ Register an EVSE with the network so it will be accessible to the rest of
        the simulation. This can only be called before any constraints have been
        registered in order to prevent dimensionality mismatch between the EVSE list
        and the

        Args:
            evse (EVSE): An EVSE object.
            voltage (float): Voltage feeding the EVSE (V).
            phase_angle (float): Phase angle of the voltage/current feeding the EVSE
                (degrees).

        Returns:
            None
        """
        # Only allow registering of EVSEs before any constraints have been added.
        if self.constraint_matrix is not None:
            raise EVSERegistrationError(
                "Attempting to register an EVSE after constraints have been added. "
                "Please register all EVSEs with the network before adding constraints."
            )
        self._EVSEs[evse.station_id] = evse
        self._voltages = np.append(self._voltages, voltage)
        self._phase_angles = np.append(self._phase_angles, phase_angle)
        # Cached information-storing objects for use by Interface.
        _ = self._update_info_store()

    def constraints_as_df(self) -> pd.DataFrame:
        """ Returns the network constraints in a pandas DataFrame.

        The index is the constraint IDs, and the columns are station IDs. The
        magnitudes (constraint limits) must be accessed separately.

        Returns:
            pd.DataFrame: The network constraints as a DataFrame.
        """
        return pd.DataFrame(
            self.constraint_matrix,
            columns=self.station_ids,
            index=self.constraint_index,
        )

    def add_constraint(
        self, current: Current, limit: float, name: Optional[str] = None
    ) -> None:
        """ Add an additional constraint to the constraint DataFrame.

        Args:
            current (Current): Aggregate current which is constrained.
                See Current for more info.
            limit (float): Upper limit on the aggregate current.
            name (str): Name of this constraint.

        Returns:
            None
        """
        if name is None:
            name = "_const_{0}".format(len(self.constraint_index))
        if name in self.constraint_index:
            warnings.warn(
                f"Constraint {name} already added. Adding input constraint as new "
                f"constraint. Use network.update_constraint to update constraint "
                f"{name}.",
                UserWarning,
            )
            name += "_v2"
        for station_id in current.index:
            if station_id not in self._EVSEs:
                raise KeyError(
                    f"Station {station_id} not found. Register station {station_id} to "
                    f"add constraint {name} to network. "
                )
        current.name = name
        self.magnitudes = np.append(self.magnitudes, limit)
        # Make a DataFrame for the constraint matrix for easy addition of the new
        # constraint
        constraint_frame: pd.DataFrame = self.constraints_as_df()
        constraint_frame = constraint_frame.append(current).fillna(0)
        # Maintain a list of constraint ids for use with constraint_current.
        self.constraint_index = list(constraint_frame.index)
        # Update the numpy matrix of constraints by reconstructing it from
        # constraint_frame.
        self.constraint_matrix = constraint_frame.reindex(
            columns=self.station_ids
        ).to_numpy()
        # Cached information-storing objects for use by Interface.
        _ = self._update_info_store()

    def remove_constraint(self, name: str) -> None:
        """ Remove a network constraint.

        Args:
            name (str): Name of constraint to remove.

        Returns:
            None
        """
        if name not in self.constraint_index:
            raise KeyError(f"Cannot remove constraint {name}: not found in network.")
        del_index: int = self.constraint_index.index(name)
        self.constraint_matrix = np.delete(self.constraint_matrix, del_index, axis=0)
        self.magnitudes = np.delete(self.magnitudes, del_index, axis=0)
        self.constraint_index.remove(name)
        # Cached information-storing objects for use by Interface.
        _ = self._update_info_store()

    def update_constraint(
        self, name: str, current: Current, limit: float, new_name: Optional[str] = None
    ) -> None:
        """ Update a network constraint with a new aggregate current, limit, and name.

        Args:
            name (str): Name of constraint to update.
            current (Current): New current to update constraint with
            limit (float): New upper limit to update constraint with
            new_name (str): New name to give constraint

        Returns:
            None
        """
        if new_name is None:
            new_name: str = name
        if name not in self.constraint_index:
            raise KeyError(f"Cannot update constraint {name}: not found in network.")
        self.remove_constraint(name)
        self.add_constraint(current, limit, name=new_name)
        # Cached information-storing objects for use by Interface.
        _ = self._update_info_store()

    def plugin(self, ev: EV, station_id: str = None) -> None:
        """ Attach EV to a specific EVSE.

        Args:
            ev (EV): EV object which will be attached to the EVSE.
            [Depreciated]
            station_id (str): ID of the EVSE.

        Returns:
            None

        Raises:
            KeyError: Raised when the station id has not yet been registered.
        """
        if station_id is not None:
            warnings.warn(
                "plugin arg 'station_id' is deprecated. Plugin now uses the value of "
                "ev.station_id directly.",
                DeprecationWarning,
            )
        if ev.station_id in self._EVSEs:
            self._EVSEs[ev.station_id].plugin(ev)
        else:
            raise KeyError("Station {0} not found.".format(ev.station_id))

    def unplug(self, station_id: str, session_id: str = None) -> None:
        """ Detach EV from a specific EVSE.

        Args:
            station_id (str): ID of the EVSE.
            session_id (str): ID of the session to be unplugged.
        Returns:
            None

        Raises:
            KeyError: Raised when the station id has not yet been registered.
        """
        if station_id in self._EVSEs:
            if session_id is None:
                warnings.warn(
                    f"Calling ChargingNetwork.unplug without a session_id argument is "
                    f"deprecated. Please include a session_id argument as this will be "
                    f"required in a future release. Unplugging EV at station "
                    f"{station_id}."
                )
                self._EVSEs[station_id].unplug()
            elif self._EVSEs[station_id].ev is None:
                warnings.warn(
                    f"Tried to remove EV with session_id {session_id} which was not "
                    f"present at station {station_id}. Found no EV instead."
                )
            elif session_id == self._EVSEs[station_id].ev.session_id:
                self._EVSEs[station_id].unplug()
            else:
                warnings.warn(
                    f"Tried to remove EV with session_id {session_id} which was not "
                    f"present at station {station_id}. Found EV with session_id "
                    f"{self._EVSEs[station_id].ev.session_id} instead."
                )
        else:
            raise KeyError("Station {0} not found.".format(station_id))

    def get_ev(self, station_id: str) -> EV:
        """ Return the EV attached to the specified EVSE.

        Args:
            station_id (str): ID of the EVSE.

        Returns:
            EV: The EV attached to the specified station.

        """
        if station_id in self._EVSEs:
            return self._EVSEs[station_id].ev
        else:
            raise KeyError("Station {0} not found.".format(station_id))

    def update_pilots(self, pilots: np.ndarray, i: int, period: float) -> None:
        """ Update the pilot signal sent to each EV. Also triggers the EVs to charge
        at the specified rate.

        Note that if a pilot is not sent to an EVSE the associated EV WILL NOT charge
        during that period. If station_id is pilots or a list does not include the
        current time index, a 0 pilot signal is passed to the EVSE. Station IDs not
        registered in the network are silently ignored.

        Args:
            pilots (np.Array): numpy array with a row for each station_id
                and a column for each time. Each entry in the Array corresponds to
                a charging rate (in A) at the station given by the row at a time given
                by the column.
            i (int): Current time index of the simulation.
            period (float): Length of the charging period. [minutes]

        Returns:
            None
        """
        ids: List[str] = self.station_ids
        for station_number in range(len(ids)):
            new_rate = pilots[station_number, i]
            self._EVSEs[ids[station_number]].set_pilot(
                new_rate, self._voltages[station_number], period
            )

    def constraint_current(
        self,
        input_schedule: np.ndarray,
        constraints: Optional[List[str]] = None,
        time_indices: Optional[List[int]] = None,
        linear: bool = False,
    ):
        """ Return the aggregate currents subject to the given constraints. If
        constraints=None, return all aggregate currents.

        Args:
            input_schedule (np.Array): 2-D matrix with each row corresponding to an
                EVSE and each column corresponding to a time index in the schedule.
            constraints (List[str]): List of constraint id's for which to calculate
                aggregate current. If None, calculates aggregate currents for all
                constraints.
            time_indices (List[int]): List of time indices for which to calculate
                aggregate current. If None, calculates aggregate currents for all
                timesteps.
            linear (bool): If True, linearize all constraints to a more conservative
                but easier to compute constraint by ignoring the phase angle and taking
                the absolute value of all load coefficients. Default False.

        Returns:
            np.ndarray: Aggregate currents subject to the given constraints.
        """
        schedule_matrix: np.ndarray = np.array(input_schedule)
        # Convert list of constraint id's to list of indices in constraint matrix
        if constraints is not None:
            constraint_indices: List[int] = [
                i
                for i in range(len(self.constraint_index))
                if self.constraint_index[i] in constraints
            ]
        else:
            constraint_indices: List[int] = list(range(len(self.constraint_index)))

        # If we only want the constraint currents at specific time indices,
        # index schedule_matrix columns using these indices
        if time_indices is not None:
            schedule_matrix = schedule_matrix[:, time_indices]

        if linear:
            return np.abs(
                self.constraint_matrix[constraint_indices] @ schedule_matrix
            ).astype("complex")
        else:
            # build vector of phase angles on EVSE
            angle_coeffs: np.ndarray = np.exp(1j * np.deg2rad(self._phase_angles))

            # multiply schedule by angles matrix element-wise
            phasor_schedule: np.ndarray = (schedule_matrix.T * angle_coeffs).T

            # multiply constraint matrix by current schedule, shifted by the phases
            return self.constraint_matrix[constraint_indices] @ phasor_schedule

    def is_feasible(
        self,
        schedule_matrix: np.ndarray,
        linear: bool = False,
        violation_tolerance: Optional[float] = None,
        relative_tolerance: Optional[float] = None,
    ) -> bool:
        """ Return if a set of current magnitudes for each load are feasible.

        For a given constraint, the larger of the violation_tolerance
        and relative_tolerance is used to evaluate feasibility.

        Args:
            schedule_matrix (np.Array): 2-D matrix with each row corresponding to
                an EVSE and each column corresponding to a time index in the schedule.

            linear (bool): If True, linearize all constraints to a more conservative
                but easier to compute constraint by ignoring the phase angle and taking
                the absolute value of all load coefficients. Default False.
            violation_tolerance (float): Absolute amount by which
                schedule_matrix may violate network constraints. Default
                None, in which case the network's violation_tolerance
                attribute is used.
            relative_tolerance (float): Relative amount by which
                schedule_matrix may violate network constraints. Default
                None, in which case the network's relative_tolerance
                attribute is used.

        Returns:
            bool: If load_currents is feasible at time t according to this set of
                constraints.
        """
        # If no violation_tolerance is specified, default to the network's
        # violation_tolerance.
        if violation_tolerance is None:
            violation_tolerance = self.violation_tolerance
        if relative_tolerance is None:
            relative_tolerance = self.relative_tolerance
        rel_magnitude_tol = self.magnitudes * relative_tolerance

        # If there are no constraints (magnitudes vector is empty) return True
        if not len(self.magnitudes):
            return True

        # Calculate aggregate currents for each constraint
        aggregate_currents = self.constraint_current(schedule_matrix, linear=linear)

        # Ensure each aggregate current is less than its limit, returning False if not
        schedule_length = schedule_matrix.shape[1]
        return np.all(
            np.tile(
                self.magnitudes + np.maximum(violation_tolerance, rel_magnitude_tol),
                (schedule_length, 1),
            ).T
            >= np.abs(aggregate_currents)
        )

    def post_charging_update(self):
        """ Hook to define actions to take after the charging update. """
        pass

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """

        attribute_dict = {}
        # Serialize non-nested attributes.
        nn_attr_lst = [
            "violation_tolerance",
            "relative_tolerance",
            "constraint_matrix",
            "magnitudes",
            "_voltages",
            "_phase_angles",
            "constraint_index",
            "_station_ids_dict",
            "max_pilot_signals",
            "min_pilot_signals",
            "is_continuous",
        ]
        for attr in nn_attr_lst:
            attribute_dict[attr] = getattr(self, attr)

        evses = {}
        for station_id, evse in self._EVSEs.items():
            # noinspection PyProtectedMember
            registry, context_dict = evse._to_registry(context_dict=context_dict)
            evses[station_id] = registry["id"]
        attribute_dict["_EVSEs"] = evses

        allowable_rates_list: List = []
        for allowable_rates_array in self.allowable_rates:
            allowable_rates_list.append(allowable_rates_array.tolist())
        attribute_dict["allowable_rates"] = allowable_rates_list

        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(
            violation_tolerance=attribute_dict["violation_tolerance"],
            relative_tolerance=attribute_dict["relative_tolerance"],
        )

        evses = {}
        for station_id, evse in attribute_dict["_EVSEs"].items():
            # noinspection PyProtectedMember
            evse_elt, loaded_dict = BaseSimObj._build_from_id(
                evse, context_dict, loaded_dict=loaded_dict
            )
            evses[station_id] = evse_elt
        out_obj._EVSEs = evses

        if attribute_dict["constraint_matrix"] is not None:
            out_obj.constraint_matrix = np.array(attribute_dict["constraint_matrix"])
        else:
            out_obj.constraint_matrix = attribute_dict["constraint_matrix"]
        out_obj.magnitudes = np.array(attribute_dict["magnitudes"])
        out_obj._voltages = np.array(attribute_dict["_voltages"])
        out_obj._phase_angles = np.array(attribute_dict["_phase_angles"])
        out_obj.constraint_index = attribute_dict["constraint_index"]

        # If the original ChargingNetwork had info stores encoded, overwrite defaults
        # with these.
        try:
            (
                _station_ids_dict,
                max_rates_list,
                min_rates_list,
                allowable_rates_list,
                is_continuous_list,
            ) = (
                attribute_dict["_station_ids_dict"],
                attribute_dict["max_pilot_signals"],
                attribute_dict["min_pilot_signals"],
                attribute_dict["allowable_rates"],
                attribute_dict["is_continuous"],
            )
        except KeyError:
            _ = out_obj._update_info_store()
        else:
            out_obj._station_ids_dict = _station_ids_dict
            out_obj.max_pilot_signals = np.array(max_rates_list)
            out_obj.min_pilot_signals = np.array(min_rates_list)
            out_obj.is_continuous = np.array(is_continuous_list, dtype="bool")
            allowable_rates_arrays: List[np.ndarray] = []
            for allowable_rates in allowable_rates_list:
                allowable_rates_arrays.append(np.array(allowable_rates))
            out_obj.allowable_rates = allowable_rates_arrays

        return out_obj, loaded_dict


class StationOccupiedError(Exception):
    """ Exception which is raised when trying to add an EV to an EVSE which is
    already occupied. """


class EVSERegistrationError(Exception):
    """ Exception which is raised when trying to add an EVSE to the network after
    constraints have already been added. """



# coding=utf-8
"""
Postprocessing functions for scheduling algorithms.
"""
import numpy as np

from acnportal.acnsim.interface import InfrastructureInfo
from acnportal.acnsim.simulator import InvalidScheduleError


def format_array_schedule(
    array_schedule: np.ndarray, infrastructure: InfrastructureInfo
):
    """ Convert schedule in array format into dictionary format.

    Args:
        array_schedule (np.array): Array of charging rates, where each
            row is the charging rates for a specific session. Rows should be in
            the same order as the station_ids attribute of infrastructure. If
            array_schedule is a 1-d array, it will be expanded into lists of
            dimension length 1 for each station.
        infrastructure (InfrastructureInfo): Description of the electrical
            infrastructure.

    Returns:
        Dict[str, List[float]]: Dictionary mapping a station_id to a list of
            charging rates.
    """
    if len(infrastructure.station_ids) != len(array_schedule):
        raise InvalidScheduleError(
            f"Proposed array_schedule has a length that is different from the number "
            f"of stations in the network."
            f"\nNumber of  stations: {len(infrastructure.station_ids)}"
            f"\nNumber of schedules: {len(array_schedule)}"
        )
    schedule = {}
    for i, station_id in enumerate(infrastructure.station_ids):
        if np.isscalar(array_schedule[i]):
            schedule[station_id] = [array_schedule[i]]
        else:
            schedule[station_id] = array_schedule[i].tolist()
    return schedule



# coding=utf-8
"""
Classes implementing different methods of estimating upper bounds on charging rate.
"""
from typing import Optional, Dict, List

import numpy as np

from acnportal.acnsim.interface import SessionInfo, Interface


class UpperBoundEstimatorBase:
    """ Abstract base class meant to be inherited from to implement new
        algorithms to estimate upper bounds on charging rate imposed by the
        EV's battery or on-board charger.

        Subclassed must implement the get_maximum_rates method.
    """

    _interface: Optional[Interface]

    def __init__(self) -> None:
        self._interface = None

    @property
    def interface(self) -> Interface:
        """ Return the algorithm's interface with the environment.

        Returns:
            Interface: An interface to the environment.

        Raises:
            ValueError: Exception raised if interface is accessed prior to an
                interface being registered.
        """
        if self._interface is not None:
            return self._interface
        else:
            raise ValueError(
                "No interface has been registered yet. Please call "
                "register_interface prior to using the algorithm."
            )

    def register_interface(self, interface: Interface) -> None:
        """ Register interface to the _simulator/physical system.
            This interface is the only connection between the algorithm and
            what it is controlling. Its purpose is to abstract the
            underlying network so that the same algorithms can run on a
            simulated environment or a physical one.

        Args:
            interface (Interface): An interface to the underlying network
                whether simulated or real.
        Returns:
            None
        """
        self._interface = interface

    def get_maximum_rates(self, sessions: List[SessionInfo]) -> Dict[str, float]:
        """ Return the maximum rate allowed for these EVs under an upper bound
        estimation algorithm.

        NOT IMPLEMENTED IN UpperBoundEstimatorBase. This method MUST be implemented in
        all subclasses.

        Args:
            sessions list(SessionInfo): List of sessions

        Returns:
            dict(str, float): Dictionary mapping session_ids to maximum
                charging rates.
        """
        raise NotImplementedError("UpperBoundEstimatorBase is an abstract class.")


class SimpleRampdown(UpperBoundEstimatorBase):
    """ Simple algorithm reclaiming unused charging capacity.

        Implements abstract class UpperBoundEstimatorBase.

        The maximum pilot is reduced whenever the actual charging rate is
        more than down_threshold lower than the pilot signal. The maximum
        pilot is increased by up_increment, whenever the actual charging rate
        is within up_threshold of the current maximum pilot signal.
    """

    up_threshold: float
    down_threshold: float
    up_increment: float
    upper_bounds: Dict[str, float]

    def __init__(
        self,
        up_threshold: float = 1,
        down_threshold: float = 1,
        up_increment: float = 1,
    ) -> None:
        super().__init__()
        self.up_threshold = up_threshold
        self.down_threshold = down_threshold
        self.up_increment = up_increment
        self.upper_bounds = {}

    def get_maximum_rates(self, sessions: List[SessionInfo]) -> Dict[str, float]:
        """ Return the maximum rate allowed for these EVs by lowering according
            to the simple rampdown algorithm.

            (See class description for algorithm details)

        Args:
            sessions list(SessionInfo): List of sessions

        Returns:
            dict(str, float): Dictionary mapping session_ids to maximum charging rates.
        """
        prev_pilot = self.interface.last_applied_pilot_signals
        prev_rate = self.interface.last_actual_charging_rate

        for session in sessions:
            if session.session_id not in self.upper_bounds:
                ub = self.interface.max_pilot_signal(session.station_id)
                self.upper_bounds[session.session_id] = ub

            # We can only apply the rampdown algorithm if we have
            # data from from the last time period of pilot signal and
            # observed charging current.
            if session.session_id in prev_pilot:
                previous_pilot = prev_pilot[session.session_id]
                previous_rate = prev_rate[session.session_id]
                ub = self.upper_bounds[session.session_id]
                if previous_pilot - previous_rate > self.down_threshold:
                    ub = previous_rate + self.up_increment
                elif ub - previous_rate < self.up_threshold:
                    ub += self.up_increment
                max_pilot = self.interface.max_pilot_signal(session.station_id)
                # ub is a float, so np.clip should return a float. The casting is to
                # satisfy the type checker.
                ub = float(np.clip(ub, a_min=0, a_max=max_pilot))
                self.upper_bounds[session.session_id] = ub
        return self.upper_bounds



# coding=utf-8
"""
This module contains methods for directly interacting with the _simulator.
"""
from copy import deepcopy
from typing import List, Union, Optional, Dict, Set, Tuple

import numpy as np
from datetime import timedelta, datetime
from collections import namedtuple
from warnings import warn

from .models import EV
from .network import ChargingNetwork


class SessionInfo:
    """ Class to store information relevant to a charging session.

        Args:
            station_id (str): Unique identifier of the station (EVSE) where the
                session takes place.
            session_id (str): Unique identifier of the charging session.
            requested_energy (float): Energy requested by the user during the
                session. [kWh]
            energy_delivered (float): Energy delivered already during the
                session. [kWh]
            arrival (int): Time index when the session begins.
            departure (int): Time index when the session ends.
            estimated_departure (int): Time index when the user estimates the session
                will end.
            current_time (int): Time index of the current time.
            min_rates (Union[float, List[float]): Lower bound for the charging
                rate of the session. If List (or np.array) length should be
                departure - arrival and each entry is a lower bound for the
                corresponding time period.
            max_rates (Union[float, List[float]): Upper bound for the charging
                rate of the session. If List (or np.array) length should be
                departure - arrival and each entry is a upper bound for the
                corresponding time period.
    """

    station_id: str
    session_id: str
    requested_energy: float
    energy_delivered: float
    arrival: int
    departure: int
    estimated_departure: int
    current_time: int
    min_rates: np.ndarray
    max_rates: np.ndarray

    def __init__(
        self,
        station_id: str,
        session_id: str,
        requested_energy: float,
        energy_delivered: float,
        arrival: int,
        departure: int,
        estimated_departure: Optional[int] = None,
        current_time: int = 0,
        min_rates: Union[float, List[float]] = 0,
        max_rates: Union[float, List[float]] = float("inf"),
    ):
        self.station_id = station_id
        self.session_id = session_id
        self.requested_energy = requested_energy
        self.energy_delivered = energy_delivered
        self.arrival = arrival
        self.departure = departure
        if self.departure <= self.arrival:
            raise ValueError(
                f"Departure must be later than arrival."
                f"\nArrival:{self.arrival}\n"
                f"Departure:{self.departure}"
            )

        if estimated_departure is None:
            self.estimated_departure = departure
        else:
            self.estimated_departure = estimated_departure

        if self.estimated_departure <= self.arrival:
            raise ValueError(
                "Departure must be later than arrival."
                f"\nArrival:{self.arrival}\n"
                f"Estimated Departure:{self.estimated_departure}"
            )

        self.current_time = current_time

        if np.isscalar(min_rates):
            self.min_rates = np.array([min_rates] * self.remaining_time)
        elif len(min_rates) == self.remaining_time:
            self.min_rates = np.array(min_rates)
        else:
            raise ValueError(
                "min_rates must be a scalar or list-like with length "
                "equal to the remaining_time of the session.\n"
                f"Length of min_rates: {len(min_rates)}\n"
                f"Remaining time: {self.remaining_time}"
            )

        if np.isscalar(max_rates):
            self.max_rates = np.array([max_rates] * self.remaining_time)
        elif len(max_rates) == self.remaining_time:
            self.max_rates = np.array(max_rates)
        else:
            raise ValueError(
                "max_rates must be a scalar or list-like with length "
                "equal to the remaining_time of the session.\n"
                f"Length of max_rates: {len(max_rates)}\n"
                f"Remaining time: {self.remaining_time}"
            )

    @property
    def remaining_demand(self) -> float:
        """ Return the Session's remaining demand in kWh."""
        return self.requested_energy - self.energy_delivered

    @property
    def arrival_offset(self) -> int:
        """ Return the time (in periods) until the arrival of the EV represented by
        this Session, or 0 if the EV has already arrived.
        """
        return max(self.arrival - self.current_time, 0)

    @property
    def remaining_time(self) -> int:
        """ Return the time remaining until the EV represented by this Session departs,
        or the time between the EV's departure and arrival if the EV has not arrived
        yet. If the EV has already departed, return 0.
        """
        remaining = min(
            self.departure - self.arrival, self.departure - self.current_time
        )
        return max(remaining, 0)


class InfrastructureInfo:
    """ Class to store information about the electrical infrastructure.

    Args:
        constraint_matrix (np.array[float]): M x N array relating the
            individual station currents to aggregate currents each of which
            is subject to a constraint. M is the number of constraints and N
            is the number of stations.
        constraint_limits (np.array[float]): Limits on each constrained link.
            Length M.
        phases (np.array[float]): Phase angle of the current at each station
            (EVSE). Length N. [deg]
        voltages (np.array[float]): Voltage of each station. Length N. [V]
        constraint_ids (List[str]): Unique identifier of each constraint.
        station_ids (List[str]): Unique identifier of each station.
        max_pilot (np.array[float]: Maximum pilot signal supported by each
            station.
        min_pilot (np.array[float]: Minimum pilot signal supported by each
            station. A non-zero min_pilot indicates that the station does
            not support any charging rates between 0 and min_pilot.  It is
            implied that all EVSEs support a pilot signal of 0, even if
            min_pilot > 0.
        allowable_pilots (List[np.array[float]): Pilot signals which each
            station supports. The allowable pilot signals for station i are
            stored in allowable_pilots[i]. If a station supports continuous
            pilot signals, the list is of length 2, where the first value is
            the lower bound on the continuous interval and the second is the
            upper bound. In continuous case, it is implied that all EVSEs
            support a pilot signal of 0, even if the continuous interval does
            not include 0.
        is_continuous (np.array[bool]): True if a station supports continuous
            pilot signals, False otherwise.
    """

    constraint_matrix: np.ndarray
    constraint_limits: np.ndarray
    phases: np.ndarray
    voltages: np.ndarray
    constraint_ids: List[str]
    station_ids: List[str]
    max_pilot: np.ndarray
    min_pilot: np.ndarray
    allowable_pilots: Union[List[np.ndarray], List[None]]
    is_continuous: np.ndarray
    _station_ids_dict: Dict[str, int]

    def __init__(
        self,
        constraint_matrix: np.ndarray,
        constraint_limits: np.ndarray,
        phases: np.ndarray,
        voltages: np.ndarray,
        constraint_ids: List[str],
        station_ids: List[str],
        max_pilot: np.ndarray,
        min_pilot: np.ndarray,
        allowable_pilots: Optional[List[np.ndarray]] = None,
        is_continuous: Optional[np.ndarray] = None,
    ):
        self.constraint_matrix = constraint_matrix
        self.constraint_limits = constraint_limits
        self.phases = phases
        self.voltages = voltages
        self.constraint_ids = constraint_ids
        self.station_ids = station_ids
        self._station_ids_dict = {
            station_id: i for i, station_id in enumerate(self.station_ids)
        }
        self.max_pilot = max_pilot
        self.min_pilot = min_pilot
        self.allowable_pilots = (
            allowable_pilots
            if allowable_pilots is not None
            else [None] * self.num_stations
        )
        # Cast an input is_continuous array to dtype bool
        if is_continuous is not None and is_continuous.dtype != "bool":
            is_continuous = is_continuous.astype("bool")
        self.is_continuous = (
            is_continuous
            if is_continuous is not None
            else np.ones(self.num_stations, dtype=bool)
        )
        self._validate()

    @property
    def num_stations(self) -> int:
        """ Return total number of stations in this network."""
        return len(self.station_ids)

    def get_station_index(self, station_id: str) -> int:
        """ Get the numerical index of a given station_id in the network."""
        return self._station_ids_dict[station_id]

    def _validate(self) -> None:
        """ Raise error if attributes do not have consistent shapes."""
        # Check number of stations
        num_stations_set: Set[int] = {
            self.constraint_matrix.shape[1],
            len(self.phases),
            len(self.voltages),
            len(self.station_ids),
            len(self.max_pilot),
            len(self.min_pilot),
            len(self.allowable_pilots),
            len(self.is_continuous),
        }
        num_constraints_set: Set[int] = {
            self.constraint_matrix.shape[0],
            len(self.constraint_limits),
            len(self.constraint_ids),
        }

        errors: List[str] = []
        if len(num_stations_set) > 1:
            errors.append(
                "Number of stations should be consistent between inputs.\n"
                "Stations implied by argument:\n"
                f"constraint_matrix: {self.constraint_matrix.shape[1]},\n"
                f"phases: {len(self.phases)},\n"
                f"voltages: {len(self.voltages)},\n"
                f"max_pilot: {len(self.max_pilot)},\n"
                f"min_pilot: {len(self.min_pilot)},\n"
                f"allowable_pilots: {len(self.allowable_pilots)},\n"
                f"is_continuous: {len(self.is_continuous)},\n"
            )
        if len(num_constraints_set) > 1:
            errors.append(
                "Number of constraints should be consistent between inputs.\n"
                "Constraints implied by argument:\n"
                f"constraint_matrix: {self.constraint_matrix.shape[0]},\n"
                f"constraint_limits: {len(self.constraint_limits)},\n"
                f"constraint_ids: {len(self.constraint_ids)},\n"
            )

        if len(errors) > 0:
            raise ValueError("\n---\n".join(errors))


Constraint = namedtuple(
    "Constraint", ["constraint_matrix", "magnitudes", "constraint_index", "evse_index"],
)


class Interface:
    """ Interface between algorithms and the ACN Simulation Environment."""

    # noinspection PyUnresolvedReferences
    _simulator: "Simulator"

    # noinspection PyUnresolvedReferences
    def __init__(self, simulator: "Simulator"):
        self._simulator = simulator

    # TODO: How do we handle default tolerances in a layer-agnostic way?
    #  Below is a proposed solution. Alternatively, InfrastructureInfo could
    #  be passed with tolerances.
    @property
    def _violation_tolerance(self) -> float:
        """ Returns the intrinsic absolute constraint violation tolerance of
        this network, for internal use by Interface.is_feasible if the algorithm does
        not provide tolerances in its call to is_feasible.

        Returns:
            float: Absolute constraint violation tolerance of the network. Submitted
                schedules may exceed constraint limits by no more than this number.
        """
        return self._simulator.network.violation_tolerance

    @property
    def _relative_tolerance(self) -> float:
        """ Returns the intrinsic relative constraint violation tolerance of
        this network, for internal use by Interface.is_feasible if the algorithm does
        not provide tolerances in its call to is_feasible.

        Returns:
            float: Relative constraint violation tolerance of the network. Submitted
                schedules may exceed constraint limits by no more than this number
                times the constraint limits.
        """
        return self._simulator.network.relative_tolerance

    @property
    def active_evs(self) -> List[EV]:
        """ Returns a list of active EVs for use by the algorithm.

        Returns:
            List[EV]: List of EVs currently plugged in and not finished.
        """
        warn(
            "Property active_evs is depreciated and will be removed in a "
            "future version. Please use active_sessions instead, "
            "as it provides a read-only copy of charging session related "
            "information."
        )
        return self._active_evs

    @property
    def _active_evs(self) -> List[EV]:
        """ Returns a list of active EVs for use by the algorithm.

        Returns:
            List[EV]: List of EVs currently plugged in and not finished.
        """
        return self._simulator.get_active_evs()

    @property
    def last_applied_pilot_signals(self) -> Dict[str, float]:
        """ Return the pilot signals that were applied in the last _iteration
            of the simulation for all active EVs.

        Does not include EVs that arrived in the current _iteration.

        Returns:
            Dict[str, float]: A dictionary with the session ID as key and the
                pilot signal as value.
        """
        i = self._simulator.iteration - 1
        if i > 0:
            return {
                ev.session_id: self._simulator.pilot_signals[
                    self._simulator.index_of_evse(ev.station_id), i
                ]
                for ev in self._active_evs
                if ev.arrival <= i
            }
        else:
            return {}

    @property
    def last_actual_charging_rate(self) -> Dict[str, float]:
        """ Return the actual charging rates in the last period for all
            active sessions.

        Returns:
            Dict[str, number]:  A dictionary with the session ID as key and actual
                charging rate as value.
        """
        return {ev.session_id: ev.current_charging_rate for ev in self._active_evs}

    @property
    def current_time(self) -> int:
        """ Get the current time (the current _iteration) of the simulator.

        Returns:
            int: The current _iteration of the simulator.
        """
        return self._simulator.iteration

    @property
    def current_datetime(self) -> datetime:
        """ Get the simulated wall time of the simulator.

        Returns:
            datetime: The datetime corresponding to the  current time step of
                the simulator.
        """
        return (
            self._simulator.start + timedelta(minutes=self.period) * self.current_time
        )

    @property
    def period(self) -> float:
        """ Return the length of each timestep in the simulation.

        Returns:
            float: Length of each time interval in the simulation. [minutes]
        """
        return self._simulator.period

    @property
    def max_recompute_time(self) -> int:
        """ Return the maximum recompute time of the simulator.

        Returns:
            int: Maximum recompute time of the simulation in number of periods.
                [periods]
        """
        return self._simulator.max_recompute

    def _active_sessions(self) -> List[SessionInfo]:
        """ Return a list of SessionInfo objects describing the currently
            charging EVs.

        Returns:
            List[SessionInfo]: List of currently active charging sessions.
        """
        return [
            SessionInfo(
                ev.station_id,
                ev.session_id,
                ev.requested_energy,
                ev.energy_delivered,
                ev.arrival,
                ev.departure,
                ev.estimated_departure,
                self.current_time,
            )
            for ev in self._active_evs
        ]

    def active_sessions(self) -> List[SessionInfo]:
        """ Return a copy of the list of SessionInfo objects describing the currently
            charging EVs.

        Returns:
            List[SessionInfo]: List of currently active charging sessions.
        """
        return deepcopy(self._active_sessions())

    def _infrastructure_info(self) -> InfrastructureInfo:
        """ Returns an InfrastructureInfo object generated from interface.

        Returns:
            InfrastructureInfo: A description of the charging infrastructure.
        """
        network: ChargingNetwork = self._simulator.network
        station_ids = network.station_ids
        max_pilot_signals = network.max_pilot_signals
        min_pilot_signals = network.min_pilot_signals
        allowable_rates = network.allowable_rates
        is_continuous = network.is_continuous
        return InfrastructureInfo(
            network.constraint_matrix,
            network.magnitudes,
            network._phase_angles,
            network._voltages,
            network.constraint_index,
            station_ids,
            max_pilot_signals,
            min_pilot_signals,
            allowable_rates,
            is_continuous,
        )

    def infrastructure_info(self) -> InfrastructureInfo:
        """ Returns a copy of the InfrastructureInfo object generated from interface.

        Returns:
            InfrastructureInfo: A description of the charging infrastructure.
        """
        return deepcopy(self._infrastructure_info())

    def allowable_pilot_signals(self, station_id: str) -> Tuple[bool, List[float]]:
        """ Returns the allowable pilot signal levels for the specified EVSE.
        One may assume an EVSE pilot signal of 0 is allowed regardless
        of this function's return values.

        Args:
            station_id (str): The ID of the station for which the allowable rates
                should be returned.

        Returns:
            bool: If the range is continuous or not
            list[float]: The sorted set of acceptable pilot signals. If continuous this
                range will have 2 values the min and the max acceptable values. [A]
        """
        infrastructure_info: InfrastructureInfo = self._infrastructure_info()
        continuity: bool = infrastructure_info.is_continuous[
            infrastructure_info.get_station_index(station_id)
        ].tolist()
        # Numpy tolist method returns "object", so type checking will fail here.
        # noinspection PyTypeChecker
        allowable_pilots: List[float] = infrastructure_info.allowable_pilots[
            infrastructure_info.get_station_index(station_id)
        ].tolist()
        return continuity, allowable_pilots

    def max_pilot_signal(self, station_id: str) -> float:
        """ Returns the maximum allowable pilot signal level for the specified EVSE.

        Args:
            station_id (str): The ID of the station.

        Returns:
            float: the maximum pilot signal supported by this EVSE. [A]
        """
        infrastructure_info: InfrastructureInfo = self._infrastructure_info()
        return infrastructure_info.max_pilot[
            infrastructure_info.get_station_index(station_id)
        ]

    def min_pilot_signal(self, station_id: str) -> float:
        """ Returns the minimum allowable pilot signal level for the EVSE.
        A zero pilot signal is always assumed to be allowed; the minimum allowable pilot
        signal returned here is the minimum nonzero pilot signal allowed by the EVSE if
        said EVSE is non-continuous.

        Args:
            station_id (str): The ID of the station.

        Returns:
            float: the minimum pilot signal supported by this EVSE. [A]
        """
        infrastructure_info: InfrastructureInfo = self._infrastructure_info()
        return infrastructure_info.min_pilot[
            infrastructure_info.get_station_index(station_id)
        ]

    def evse_voltage(self, station_id: str) -> float:
        """ Returns the voltage of the EVSE.

        Args:
            station_id (str): The ID of the station.

        Returns:
            float: voltage of the EVSE. [V]
        """
        infrastructure_info: InfrastructureInfo = self._infrastructure_info()
        return infrastructure_info.voltages[
            infrastructure_info.get_station_index(station_id)
        ]

    def evse_phase(self, station_id: str) -> float:
        """ Returns the phase angle of the EVSE.

        Args:
            station_id (str): The ID of the station.

        Returns:
            float: phase angle of the EVSE. [degrees]
        """
        infrastructure_info: InfrastructureInfo = self._infrastructure_info()
        return infrastructure_info.phases[
            infrastructure_info.get_station_index(station_id)
        ]

    def remaining_amp_periods(self, ev: SessionInfo) -> float:
        """ Return the EV's remaining demand in A*periods.

        Args:
            ev (SessionInfo): The SessionInfo object for which to get remaining demand.

        Returns:
            float: the EV's remaining demand in A*periods.
        """
        return self._convert_to_amp_periods(ev.remaining_demand, ev.station_id)

    def _convert_to_amp_periods(self, kwh: float, station_id: str) -> float:
        """ Convert the given energy in kWh to A*periods based on the voltage
            at EVSE station_id.

        Returns:
            float: kwh in A*periods.

        """
        return kwh * 1000 / self.evse_voltage(station_id) * 60 / self.period

    def get_constraints(self) -> Constraint:
        """ Get the constraint matrix and corresponding EVSE ids for the network.

        Returns:
            Constraint: namedtuple including the following attributes:
                constraint_matrix (np.ndarray): Matrix representing the constraints
                    of the network. Each row is a constraint and each
                    column is an index.
                constraint_limits (np.ndarray): Vector of bounding limits for each
                    constraint (1 for each constraint).
                constraint_ids (List[str]): Names of each constraint.
                station_ids (List[str]): Names of each station.

        """
        infrastructure_info: InfrastructureInfo = self._infrastructure_info()
        return Constraint(
            infrastructure_info.constraint_matrix,
            infrastructure_info.constraint_limits,
            infrastructure_info.constraint_ids,
            infrastructure_info.station_ids,
        )

    def is_feasible(
        self,
        load_currents: Dict[str, List[float]],
        linear: bool = False,
        violation_tolerance: Optional[float] = None,
        relative_tolerance: Optional[float] = None,
    ) -> bool:
        # TODO: Should Interface.is_feasible replace network is_feasible?
        # TODO: ACN-Live should not accept violation_tolerance, relative_tolerance
        #  args that are less than the built-in network tols.
        """ Return if a set of current magnitudes for each load are feasible.

        Wraps Network's is_feasible method.

        For a given constraint, the larger of the violation_tolerance
        and relative_tolerance is used to evaluate feasibility.

        Args:
            load_currents (Dict[str, List[number]]): Dictionary mapping load_ids to
                schedules of charging rates.
            linear (bool): If True, linearize all constraints to a more conservative
                but easier to compute constraint by ignoring the phase angle and taking
                the absolute value of all load coefficients. Default False.
            violation_tolerance (float): Absolute amount by which
                schedule may violate network constraints. Default
                None, in which case the network's violation_tolerance
                attribute is used.
            relative_tolerance (float): Relative amount by which
                schedule may violate network constraints. Default
                None, in which case the network's relative_tolerance
                attribute is used.

        Returns:
            bool: If load_currents is feasible at time t according to this set of
                constraints.
        """
        if violation_tolerance is None:
            violation_tolerance = self._violation_tolerance
        if relative_tolerance is None:
            relative_tolerance = self._relative_tolerance

        if len(load_currents) == 0:
            return True

        # Check that all schedules are the same length
        schedule_lengths = set(len(x) for x in load_currents.values())
        if len(schedule_lengths) > 1:
            raise InvalidScheduleError("All schedules should have the same length.")
        schedule_length = schedule_lengths.pop()

        # Convert input schedule into its matrix representation
        schedule_matrix = np.array(
            [
                load_currents[evse_id]
                if evse_id in load_currents
                else [0] * schedule_length
                for evse_id in self._simulator.network.station_ids
            ]
        )
        return self._simulator.network.is_feasible(
            schedule_matrix, linear, violation_tolerance, relative_tolerance
        )

    def get_prices(self, length: int, start: Optional[int] = None) -> np.ndarray:
        """ Get a vector of prices beginning at time start and continuing for length
        periods. ($/kWh)

        Args:
            length (int): Number of elements in the prices vector. One entry per period.
            start (int): Time step of the simulation where price vector should begin.
                If None, uses the current timestep of the simulation. Default None.

        Returns:
            np.ndarray[float]: Array of floats where each entry is the price for the
                corresponding period. ($/kWh)
        """
        if "tariff" in self._simulator.signals:
            if start is None:
                start = self.current_time
            price_start = self._simulator.start + timedelta(minutes=self.period) * start
            return np.array(
                self._simulator.signals["tariff"].get_tariffs(
                    price_start, length, self.period
                )
            )
        else:
            raise ValueError("No pricing method is specified.")

    def get_demand_charge(self, start: Optional[int] = None) -> float:
        """ Get the demand charge for the given period. ($/kW)

        Args:
            start (int): Time step of the simulation where price vector should begin.
                If None, uses the current timestep of the simulation. Default None.

        Returns:
            float: Demand charge for the given period. ($/kW)
        """
        if "tariff" in self._simulator.signals:
            if start is None:
                start = self.current_time
            price_start = self._simulator.start + timedelta(minutes=self.period) * start
            return self._simulator.signals["tariff"].get_demand_charge(price_start)
        else:
            raise ValueError("No pricing method is specified.")

    def get_prev_peak(self) -> float:
        """ Get the highest aggregate peak demand so far in the simulation.

        Returns:
            float: Peak demand so far in the simulation. (A)
        """
        return self._simulator.peak


class InvalidScheduleError(Exception):
    """ Raised when the schedule passed to the simulator is invalid. """



import json
import os.path
from datetime import datetime, timedelta
from decimal import Decimal
from copy import copy
from typing import List


class TariffSchedule(object):
    """ Container to store individual tariff schedules.

    Args:
        doc (dict): A tariff schedule json document as a dict.
    Attributes:
        id (str): Identifier of the schedule.
        start (int): (Month, Day) when the schedule takes effect.
        end (int): (Month, Day) of the last day when the schedule is in effect.
        dow_mask (List[bool]): Boolean mask which represents which days of the week a schedule is valid for. 0 index is
            Monday, 6 is Sunday.
        tariffs (List[Tuple[float, float]]): List of time, tariff pairs where time is measured in hours since midnight
            and tariffs are in $/kWh.
    """

    def __init__(self, doc):
        self.id = doc["id"]
        self.start = tuple(int(x) for x in doc["effective_start"].split("-"))
        self.end = tuple(int(x) for x in doc["effective_end"].split("-"))
        if doc["dow_mask"] == "WEEKDAYS":
            self.dow_mask = [True] * 5 + [False] * 2
        elif doc["dow_mask"] == "WEEKENDS":
            self.dow_mask = [False] * 5 + [True] * 2
        elif doc["dow_mask"] == "ALL":
            self.dow_mask = [True] * 7
        else:
            raise ValueError("dow_mask must be WEEKEDAY, WEEKENDS, or ALL.")
        self.tariffs = [
            (Decimal(doc["times"][i]), float(doc["tariffs"][i]))
            for i in range(len(doc["times"]))
        ]
        self.tariffs.sort()
        if self.tariffs[0][0] != 0:
            raise ValueError(
                "Error with tariff schedule. "
                "Schedule must start at time 0, began with time {0}".format(
                    self.tariffs[0][0]
                )
            )
        self.demand_charge = doc["demand_charge"]


class TimeOfUseTariff(object):
    """ Class to represent the time-of-use tariff schedule of a utility.

    Args:
        filename (str): Name of the tariff structure file excluding its extension which should be .json.
        tariff_dir (str): Path to the directory where tariff files are stored. Defaults to the pre-installed tariff
            directory included with the package.

    Attributes:
        name (str): Name of the tariff.
        effective (str): Date when the tariff went into effect. Useful if a tariff has been modified but kept the same
            name.
    """

    def __init__(self, filename, tariff_dir="tariff_schedules"):
        with open(
            os.path.join(os.path.dirname(__file__), tariff_dir, filename + ".json")
        ) as f:
            tariff_file = json.load(f)
        self.name = tariff_file["name"]
        self.effective = tariff_file["effective"]
        self._schedule = [TariffSchedule(s) for s in tariff_file["schedule"]]
        to_add = []
        for s in self._schedule:
            if s.end < s.start:
                s_copy = copy(s)
                s_copy.start = (1, 1)
                to_add.append(s_copy)
                s.end = (12, 31)
        self._schedule.extend(to_add)
        self._schedule.sort(key=lambda x: x.start)

    def _get_tariff_schedule(self, date_time: datetime) -> TariffSchedule:
        """ Return the tariff schedule in effect for the given datetime.

        Args:
            date_time (datetime): A datetime object.

        Returns:
            TariffSchedule: An object representing the tariff schedule for a given time.
        """
        valid_schedules = [
            s
            for s in self._schedule
            if s.dow_mask[date_time.weekday()]
            and s.start <= (date_time.month, date_time.day) <= s.end
        ]
        if len(valid_schedules) == 0:
            raise ValueError("No valid tariff schedule for {0}".format(date_time))
        elif len(valid_schedules) > 1:
            raise ValueError(
                "More than one tariff schedule is valid for {0}".format(date_time)
            )
        else:
            return valid_schedules[0]

    def get_tariff(self, date_time: datetime) -> float:
        """ Return the tariff in effect at a given datetime.

        Args:
            date_time (datetime): A datetime object.

        Returns:
            float: Tariff [$/kWh]
        """
        tariff_schedule = self._get_tariff_schedule(date_time)
        target_hour = (
            Decimal(date_time.hour)
            + Decimal(date_time.minute) / 60
            + Decimal(date_time.second) / 3600
        )
        for r in sorted(tariff_schedule.tariffs, reverse=True):
            if target_hour >= r[0]:
                return r[1]
        raise ValueError(
            "Error with tariff schedule. Could not find a valid price for {0}.".format(
                date_time
            )
        )

    def get_tariffs(self, start: datetime, length: int, period: int) -> List[float]:
        """ Return a list of tariffs beginning at time start and continuing for length time intervals.


        Args:
            start (datetime): Initial time when the tariff list should begin.
            length (int): Number of time intervals to cover.
            period (int): Length of each time interval in minutes.

        Returns:
            List[float]: A list of tariffs, one for each time interval. [$/kWh]
        """
        return [
            self.get_tariff(start + t * timedelta(minutes=period))
            for t in range(length)
        ]

    def get_demand_charge(self, date_time: datetime) -> float:
        """ Return the demand charge for rates at the given time.

        Args:
            date_time (datetime): A datetime object.

        Returns:
            float: Tariff [$/kW]
        """
        tariff_schedule = self._get_tariff_schedule(date_time)
        return tariff_schedule.demand_charge



# coding=utf-8
"""
Integration tests using gym_acnsim.
"""
from typing import List, Dict
from unittest import TestCase

from acnportal import acnsim
from acnportal.acnsim import Simulator, Interface, EV
from acnportal.acnsim import acndata_events
from acnportal.acnsim import sites
from acnportal.algorithms import BaseAlgorithm
from datetime import datetime

import pytz
import numpy as np
import os
import json
from copy import deepcopy


class EarliestDeadlineFirstAlgoStateful(BaseAlgorithm):
    """ See EarliestDeadlineFirstAlgo in tutorial 2. This is a stateful version that
    occasionally records charging rates and pilot signals to test the last_applied_pilot_signals
    and last_actual_charging_rate functions in Interface.
    """

    def __init__(self, increment=1):
        super().__init__()
        self._increment = increment
        self.polled_pilots = {}
        self.polled_charging_rates = {}
        self.max_recompute = 1

    def schedule(self, active_evs: List[EV]) -> Dict[str, List[float]]:
        """ Schedule EVs by first sorting them by departure time, then
        allocating them their maximum feasible rate.

        Implements abstract method schedule from BaseAlgorithm.

        See class documentation for description of the algorithm.

        This version stores the current pilot signals and charging rates
        every 100 timesteps.

        Args:
            active_evs (List[EV]): see BaseAlgorithm

        Returns:
            Dict[str, List[float]]: see BaseAlgorithm
        """
        schedule = {ev.station_id: [0] for ev in active_evs}

        sorted_evs = sorted(active_evs, key=lambda x: x.departure)

        for ev in sorted_evs:
            schedule[ev.station_id] = [self.interface.max_pilot_signal(ev.station_id)]

            while not self.interface.is_feasible(schedule):
                schedule[ev.station_id][0] -= self._increment

                if schedule[ev.station_id][0] < 0:
                    schedule[ev.station_id] = [0]
                    break
        if not self.interface.current_time % 100:
            self.polled_pilots[
                str(self.interface.current_time)
            ] = self.interface.last_applied_pilot_signals
            self.polled_charging_rates[
                str(self.interface.current_time)
            ] = self.interface.last_actual_charging_rate
        return schedule


def to_array_dict(list_dict: Dict[str, List[float]]) -> Dict[str, np.ndarray]:
    """ Converts a dictionary of strings to lists to a dictionary of
    strings to numpy arrays.
    """
    return {key: np.array(value) for key, value in list_dict.items()}


class TestIntegration(TestCase):
    # noinspection PyMissingOrEmptyDocstring
    @classmethod
    def setUpClass(cls) -> None:
        timezone = pytz.timezone("America/Los_Angeles")
        start = timezone.localize(datetime(2018, 9, 5))
        end = timezone.localize(datetime(2018, 9, 6))
        period = 5  # minute
        voltage = 220  # volts
        default_battery_power = 32 * voltage / 1000  # kW
        site = "caltech"

        cn = sites.caltech_acn(basic_evse=True, voltage=voltage)

        api_key = "DEMO_TOKEN"
        events = acndata_events.generate_events(
            api_key, site, start, end, period, voltage, default_battery_power
        )

        cls.sch = EarliestDeadlineFirstAlgoStateful()
        required_interface = Interface

        cls.sim = Simulator(
            deepcopy(cn),
            cls.sch,
            deepcopy(events),
            start,
            period=period,
            verbose=False,
            interface_type=required_interface,
        )
        cls.sim.run()

        with open(
            os.path.join(
                os.path.dirname(__file__), "edf_algo_true_analysis_fields.json"
            ),
        ) as infile:
            cls.edf_algo_true_analysis_dict = json.load(infile)

        with open(
            os.path.join(
                os.path.dirname(__file__), "edf_algo_true_datetimes_array.json"
            )
        ) as infile:
            cls.edf_algo_true_datetimes_array = json.load(infile)

        with open(
            os.path.join(os.path.dirname(__file__), "edf_algo_true_info_fields.json")
        ) as infile:
            cls.edf_algo_true_info_dict = json.load(infile)

    def test_aggregate_current(self) -> None:
        np.testing.assert_allclose(
            acnsim.aggregate_current(self.sim),
            np.array(self.edf_algo_true_analysis_dict["aggregate_current"]),
        )

    def _compare_array_dicts(self, array_dict1, array_dict2) -> None:
        self.assertEqual(
            sorted(list(array_dict1.keys())), sorted(list(array_dict2.keys()))
        )
        for key in array_dict1.keys():
            np.testing.assert_allclose(array_dict1[key], array_dict2[key])

    def test_constraint_currents_all_magnitudes(self) -> None:
        self._compare_array_dicts(
            acnsim.constraint_currents(self.sim),
            to_array_dict(
                self.edf_algo_true_analysis_dict["constraint_currents_all_linear"]
            ),
        )

    def test_constraint_currents_some_magnitudes(self) -> None:
        self._compare_array_dicts(
            acnsim.constraint_currents(
                self.sim, constraint_ids=["Primary A", "Secondary C"]
            ),
            to_array_dict(
                self.edf_algo_true_analysis_dict["constraint_currents_some_linear"]
            ),
        )

    def test_proportion_of_energy_delivered(self) -> None:
        self.assertEqual(
            acnsim.proportion_of_energy_delivered(self.sim),
            self.edf_algo_true_analysis_dict["proportion_of_energy_delivered"],
        )

    def test_proportion_of_demands_met(self) -> None:
        self.assertEqual(
            acnsim.proportion_of_demands_met(self.sim),
            self.edf_algo_true_analysis_dict["proportion_of_demands_met"],
        )

    def test_current_unbalance_nema_error(self) -> None:
        with self.assertRaises(ValueError):
            acnsim.current_unbalance(
                self.sim, ["Primary A", "Primary B", "Primary C"], unbalance_type="ABC"
            )

    def test_current_unbalance_nema_warning(self) -> None:
        with self.assertWarns(DeprecationWarning):
            acnsim.current_unbalance(
                self.sim, ["Primary A", "Primary B", "Primary C"], type="NEMA"
            )

    def test_current_unbalance_nema(self) -> None:
        # A RuntimeWarning is expected to be raised in this test case as
        # of acnportal v.1.0.3. See Github issue #57 for a discussion of
        # why this occurs.
        with self.assertWarns(RuntimeWarning):
            np.testing.assert_allclose(
                acnsim.current_unbalance(
                    self.sim, ["Primary A", "Primary B", "Primary C"]
                ),
                np.array(
                    self.edf_algo_true_analysis_dict["primary_current_unbalance_nema"]
                ),
                atol=1e-6,
            )
        with self.assertWarns(RuntimeWarning):
            np.testing.assert_allclose(
                acnsim.current_unbalance(
                    self.sim, ["Secondary A", "Secondary B", "Secondary C"]
                ),
                np.array(
                    self.edf_algo_true_analysis_dict["secondary_current_unbalance_nema"]
                ),
                atol=1e-6,
            )

    def test_datetimes_array_tutorial_2(self) -> None:
        np.testing.assert_equal(
            acnsim.datetimes_array(self.sim),
            np.array(
                [
                    np.datetime64(date_time)
                    for date_time in self.edf_algo_true_datetimes_array
                ]
            ),
        )

    def test_tutorial_2(self) -> None:
        old_evse_keys = list(self.edf_algo_true_info_dict["pilot_signals"].keys())
        new_evse_keys = self.sim.network.station_ids
        self.assertEqual(sorted(new_evse_keys), sorted(old_evse_keys))

        edf_algo_new_info_dict = {
            field: self.sim.__dict__[field]
            for field in self.edf_algo_true_info_dict.keys()
        }
        edf_algo_new_info_dict["charging_rates"] = {
            self.sim.network.station_ids[i]: list(
                edf_algo_new_info_dict["charging_rates"][i]
            )
            for i in range(len(self.sim.network.station_ids))
        }
        edf_algo_new_info_dict["pilot_signals"] = {
            self.sim.network.station_ids[i]: list(
                edf_algo_new_info_dict["pilot_signals"][i]
            )
            for i in range(len(self.sim.network.station_ids))
        }

        for evse_key in new_evse_keys:
            len_pilots: int = len(
                self.edf_algo_true_info_dict["pilot_signals"][evse_key]
            )
            true_ps: np.ndarray = np.array(
                self.edf_algo_true_info_dict["pilot_signals"][evse_key]
            )
            curr_ps: np.ndarray = np.array(
                edf_algo_new_info_dict["pilot_signals"][evse_key]
            )
            np.testing.assert_allclose(true_ps, curr_ps[:len_pilots])

            len_cr: int = len(self.edf_algo_true_info_dict["charging_rates"][evse_key])
            true_cr: np.ndarray = np.array(
                self.edf_algo_true_info_dict["charging_rates"][evse_key]
            )
            curr_cr: np.ndarray = np.array(
                edf_algo_new_info_dict["charging_rates"][evse_key]
            )
            np.testing.assert_allclose(true_cr, curr_cr[:len_cr])

        self.assertEqual(
            edf_algo_new_info_dict["peak"], self.edf_algo_true_info_dict["peak"]
        )

    def test_lap_interface_func(self) -> None:
        with open(
            os.path.join(os.path.dirname(__file__), "edf_algo_pilot_signals.json")
        ) as infile:
            self.edf_algo_true_lap = json.load(infile)

        self.assertDictEqual(self.edf_algo_true_lap, self.sch.polled_pilots)

    def test_cr_interface_func(self) -> None:
        with open(
            os.path.join(os.path.dirname(__file__), "edf_algo_charging_rates.json")
        ) as infile:
            self.edf_algo_true_cr = json.load(infile)

        self.assertDictEqual(self.edf_algo_true_cr, self.sch.polled_charging_rates)


class EmptyScheduler(BaseAlgorithm):
    """
    Always submits an empty schedule (empty dict) as the output of
    its run function.
    """

    def __init__(self) -> None:
        super().__init__()

    def schedule(self, active_evs) -> Dict[str, List[float]]:
        """
        Return an empty schedule.

        Implements abstract method schedule from BaseAlgorithm.

        Args:
            active_evs (List[EV]): see BaseAlgorithm

        Returns:
            Dict[str, List[float]]: see BaseAlgorithm
        """
        return {}


class TestEmptyScheduleSim(TestCase):
    # noinspection PyMissingOrEmptyDocstring
    @classmethod
    def setUpClass(cls) -> None:
        cls.plugin_event1 = acnsim.PluginEvent(
            10, acnsim.EV(10, 20, 30, "PS-001", "EV-001", acnsim.Battery(100, 50, 20))
        )
        cls.plugin_event2 = acnsim.PluginEvent(
            20, acnsim.EV(20, 30, 40, "PS-002", "EV-002", acnsim.Battery(100, 50, 20))
        )

        cls.evse1 = acnsim.EVSE("PS-001", max_rate=32)
        cls.evse2 = acnsim.EVSE("PS-002", max_rate=32)

        cls.event_queue = acnsim.EventQueue()
        cls.event_queue.add_events([cls.plugin_event1, cls.plugin_event2])

        cls.network = acnsim.ChargingNetwork()
        cls.network.register_evse(cls.evse1, 220, 30)
        cls.network.register_evse(cls.evse2, 220, 30)

        # Simulator with scheduler that always returns an empty
        # schedule.
        cls.simulator = acnsim.Simulator(
            cls.network,
            EmptyScheduler(),
            cls.event_queue,
            datetime(2019, 1, 1),
            verbose=False,
        )

        cls.simulator.run()

    def test_pilot_signals_empty_schedule(self) -> None:
        np.testing.assert_allclose(self.simulator.pilot_signals, np.zeros((2, 31)))

    def test_charging_rates_empty_schedule(self) -> None:
        np.testing.assert_allclose(self.simulator.charging_rates, np.zeros((2, 31)))



import math
from datetime import datetime

from ..models.ev import EV
from ..models.battery import Battery
from . import PluginEvent
from .event_queue import EventQueue
from acnportal.acndata import DataClient


def generate_events(token, site, start, end, period, voltage, max_rate, **kwargs):
    """ Return EventQueue filled using events gathered from the acndata API.

    Args:
        See get_evs().

    Returns:
        EventQueue: An EventQueue filled with Events gathered through the acndata API.

    """
    evs = get_evs(token, site, start, end, period, voltage, max_rate, **kwargs)
    events = [PluginEvent(sess.arrival, sess) for sess in evs]
    return EventQueue(events)


def get_evs(
    token,
    site,
    start: datetime,
    end: datetime,
    period,
    voltage,
    max_battery_power,
    max_len=None,
    battery_params=None,
    force_feasible=False,
):
    """ Return a list of EVs gathered from the acndata API.

    Args:
        token (str): API token needed to access the acndata API.
        site (str): ACN id for the site where data should be gathered.
        start (datetime): Only return sessions which began after start.
        end (datetime): Only return session which began before end.
        period (int): Length of each time interval. (minutes)
        voltage (float): Voltage of the network.
        max_battery_power (float): Default maximum charging power for batteries.
        max_len (int): Maximum length of a session. (periods) Default None.
        battery_params (Dict[str, object]): Dictionary containing parameters for the EV's battery. Three keys are
            supported. If none, Battery type is used with default configuration. Default None.
            - 'type' maps to a Battery-like class. (required)
            - 'capacity_fn' maps to a function which takes in the the energy delivered to the car, the length of the
                session, the period of the simulation, and the voltage of the system. It should return a tuple with
                the capacity of the battery and the initial charge of the battery both in A*periods.
            - 'kwargs' maps to a dictionary of keyword arguments which will be based to the Battery constructor.
        force_feasible (bool): If True, the requested_energy of each session will be reduced if it exceeds the amount
            of energy which could be delivered at maximum rate during the duration of the charging session.
            Default False.
    Returns:

    """
    client = DataClient(token)
    docs = client.get_sessions_by_time(site, start, end)
    evs = []
    offset = _datetime_to_timestamp(start, period)
    for d in docs:
        evs.append(
            _convert_to_ev(
                d,
                offset,
                period,
                voltage,
                max_battery_power,
                max_len,
                battery_params,
                force_feasible,
            )
        )
    return evs


def _convert_to_ev(
    d,
    offset,
    period,
    voltage,
    max_battery_power,
    max_len=None,
    battery_params=None,
    force_feasible=False,
):
    """ Convert a json document for a single charging session from acndata into an EV object.

    Args:
        d (dict): Session expressed as a dictionary. See acndata API for more details.
        offset (int): Simulation timestamp of the beginning of the simulation.
        See get_evs() for additional args.

    Returns:
        EV: EV object with data from the acndata session doc.
    """
    arrival = _datetime_to_timestamp(d["connectionTime"], period) - offset
    departure = _datetime_to_timestamp(d["disconnectTime"], period) - offset

    if max_len is not None and departure - arrival > max_len:
        departure = arrival + max_len

    # requested_energy = d['kWhDelivered'] * 1000 * (60 / period) / voltage  # A*periods

    if force_feasible:
        delivered_energy = min(
            d["kWhDelivered"], max_battery_power * (departure - arrival) * (period / 60)
        )
    else:
        delivered_energy = d["kWhDelivered"]

    session_id = d["sessionID"]
    station_id = d["spaceID"]

    if battery_params is None:
        battery_params = {"type": Battery}
    batt_kwargs = battery_params["kwargs"] if "kwargs" in battery_params else {}
    if "capacity_fn" in battery_params:
        cap, init = battery_params["capacity_fn"](
            delivered_energy, departure - arrival, voltage, period
        )
    else:
        cap = delivered_energy
        init = 0
    batt = battery_params["type"](cap, init, max_battery_power, **batt_kwargs)

    # delivered_energy_amp_periods = delivered_energy * 1000 * (60 / period) / voltage
    return EV(arrival, departure, delivered_energy, station_id, session_id, batt)


def _datetime_to_timestamp(dt, period, round_up=False):
    """ Convert a datetime object to a timestamp measured in simulation periods.

    Args:
        dt (datetime): Datetime to be converted to a simulation timestamp.
        period (int): Length of one time interval in the simulation. (minutes)
        round_up (bool): If True, round up when casting timestamp to int, else round down.

    Returns:
        int: dt expressed as a simulation timestamp.
    """
    ts = dt.timestamp() / (60 * period)
    if round_up:
        return int(math.ceil(ts))
    else:
        return int(ts)



# coding=utf-8
"""
This module contains a base class shared by all ACN-Sim objects.
"""
import json
import operator
import os
from typing import Optional, Dict, Any, Tuple

import numpy as np

# noinspection PyProtectedMember
from pydoc import locate
import warnings
import pkg_resources
import pandas
from pandas.io.common import stringify_path, get_handle, get_filepath_or_buffer
import numpy

__NOT_SERIALIZED_FLAG__ = "__NOT_SERIALIZED__"


# Uses methodology discussed in https://stackoverflow.com/a/16372436
# by Martijn Pieters.


def _operator_error_hooks(cls):
    operator_hooks = [
        name for name in dir(operator) if name.startswith("__") and name.endswith("__")
    ]

    def add_hook(name):
        # noinspection PyUnusedLocal
        def op_hook(*args, **kwargs):
            raise TypeError(
                f"This object is a stub object whose methods are not "
                f"callable. Call {name} after correctly instantiating"
                f" this object."
            )

        try:
            setattr(cls, name, op_hook)
        except (AttributeError, TypeError):
            pass

    for hook_name in operator_hooks:
        add_hook(hook_name)
    return cls


@_operator_error_hooks
class ErrorAllWrapper:
    """
    This wrapper class wraps a string representing an object that acnsim
     does not know how to
    serialize. If any operator or method is called on the wrapped
    string, an error is raised. The only attribute accessible is the
    data attribute, which returns the wrapped string.
    """

    def __init__(self, data):
        self._data = data

    def __getattr__(self, item):
        if item == "data":
            return
        raise TypeError(
            f"This object is a stub object whose methods are not "
            f"callable. Call {item} after correctly instantiating this "
            f"object."
        )

    @property
    def data(self):
        return self._data


class NpEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=E0202
        if isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.ndarray):
            return o.tolist()
        else:
            json.JSONEncoder.default(self, o)


class BaseSimObj:
    """
    Base class for all ACN-Sim objects. Includes functions for
    representation and serialization of ACN-Sim objects.
    """

    def __repr__(self):
        """
        General string representation of an ACN-Sim object. Unless they
        are non-iterable builtins, attributes' default (object) repr
        functions are used.

        Returns:
            str: A representation of the object in the following form:
                [module name].[class name](attr1=[value1],
                                           attr2=[value2],
                                           ...)
        """
        attr_repr_lst = []
        for key, value in self.__dict__.items():
            if value.__class__.__module__ == "builtins":
                try:
                    _ = iter(value)
                except TypeError:
                    attr_repr_lst.append(f"{key}={value}")
                else:
                    attr_repr_lst.append(f"{key}={object.__repr__(value)}")
            else:
                attr_repr_lst.append(f"{key}={object.__repr__(value)}")
        attr_repr = ", ".join(attr_repr_lst)
        return f"{self.__module__}.{self.__class__.__name__}({attr_repr})"

    @staticmethod
    def _none_to_empty_dict(*args):
        """
        Returns a new args list that replaces each None arg with {}.
        """
        out_arg_lst = []
        for arg in args:
            if arg is None:
                out_arg_lst.append({})
            else:
                out_arg_lst.append(arg)
        return out_arg_lst

    def to_json(self, path_or_buf=None):
        """ Returns a JSON string representing self.
        Currently, only non-compressed file types are supported as the
        output file type.

        Args:
            path_or_buf (FilePathOrBuffer): File path or object. If not
            specified, the result is returned as a string.
        """
        # The code here is from pandas 1.0.1, io.json.to_json(), with
        # modifications.
        json_serializable_data = self._to_registry()[0]
        if json_serializable_data["version"] is None:
            warnings.warn(
                f"Missing a recorded version of acnportal in the dict "
                f"representation. Loading will not run an acnportal "
                f"version check.",
                UserWarning,
            )
        if json_serializable_data["dependency_versions"] is None:
            warnings.warn(
                f"Missing recorded versions of dependencies in the "
                f"dict representation. Loading will not run a "
                f"dependency version check.",
                UserWarning,
            )
        path_or_buf = stringify_path(path_or_buf)
        if isinstance(path_or_buf, str):
            fh, _ = get_handle(path_or_buf, "w")
            try:
                json.dump(json_serializable_data, fh, cls=NpEncoder)
                # Add a newline to the EOF.
                fh.write("\n")
            finally:
                fh.close()
        elif path_or_buf is None:
            return json.dumps(json_serializable_data, cls=NpEncoder)
        else:
            json.dump(json_serializable_data, path_or_buf, cls=NpEncoder)
            # Add a newline to the EOF.
            path_or_buf.write("\n")

    def _to_registry(self, context_dict=None):
        """
        Returns a JSON serializable representation of self.

        The serializer is invoked using `obj.to_json()`, but the
        conversion into a serializable representation occurs in this
        method.

        This method is protected (i.e. underscored) so users do not
        directly use this method. PyNoInspection comments have been
        added where this function is used within the package.

        Serializer behaviors:

        - Any native ACN-Sim object (that is, any unaltered object
        provided in `acnsim`) dumps without error and with
        accurate preservation of attributes.

        - An extension of an ACN-Sim object that appropriately defines a
        `_to_dict` method will dump without error and with
        accurate attribute preservation. This also applies to ACN-Sim
        objects that have as attributes extensions of ACN-Sim objects
        with defined `_to_dict` methods.

        - An extension that doesn't define `_to_dict` will dump without
        error, but with partial preservation of attributes. Objects that
        do not inherit from BaseSimObj and either are not themselves
        JSON serializable or contain attributes that are not JSON
        serializable will not be serialized.

        - A warning is thrown for each attribute not explicitly handled
        in the serialization.

        - A warning is thrown for each attribute that is neither naively
        serializable nor handled by any object's `_to_registry` method.

        The serializer returns a dict with two keys. An `id`, and a
        `context_dict` that maps ID's to JSON serializable
        representations of  objects. This ensures that objects appearing
        multiple times in another object's attributes are not re-loaded
        as different objects. For example, a Simulator object may have
        an EV appearing in both the `ev_history`, and the
        `event_history`, as part of a `PluginEvent`. The `id`-based
        serialization ensures that on loading, the same `EV` object
        occupies both the `ev_history` and the `event_history`.

        As serialization of nested objects occurs recursively, this
        method also optionally accepts a `context_dict` that contains
        objects (via the ID's) that have already been converted into a
        serializable form, to which the serializable form of this object
        is added.

        Args:
            context_dict (Dict[str, JSON Serializable]): Dict mapping
                object ID's to object JSON serializable representations.

        Returns:
            Dict[str, JSON Serializable]: A JSON serializable
                representation of this object. Takes the form:
                    ```
                    {'id': [This object's id],
                     'context_dict': [The dict mapping all object id's
                                      to their JSON serializable
                                      representations],
                     'version': [commit hash of acnportal at creation
                                 of this object]}
                    ```
            Dict[str, JSON Serializable]: For clarity, the context_dict
                is returned since it is modified by this function.

        Warns:
            UserWarning: If any attributes are present in the object
                not handled by the object's `_to_dict` method.
            UserWarning: If any of the unhandled attributes are not
                JSON serializable.

        """
        first_call = context_dict is None
        (context_dict,) = self._none_to_empty_dict(context_dict)
        obj_id = f"{id(self)}"

        # Check if this object has already been converted, and return
        # the appropriate dict if this is the case.
        if obj_id in context_dict:
            return {"id": obj_id, "context_dict": context_dict}, context_dict

        obj_type = f"{self.__module__}.{self.__class__.__name__}"

        # Get a dictionary of the attributes of this object using the
        # object's _to_dict method.
        attribute_dict, context_dict = self._to_dict(context_dict)

        # Check that all attributes have been serialized.
        # Warn if some attributes weren't serialized.
        if attribute_dict.keys() != self.__dict__.keys():
            unserialized_keys = set(self.__dict__.keys()) - set(attribute_dict.keys())
            warnings.warn(
                f"Attributes {unserialized_keys} present in object of "
                f"type {obj_type} but not handled by object's _to_dict "
                f"method. Serialized object may not load correctly. "
                f"Write a _to_dict method and re-dump, or write an "
                f"appropriate _from_dict method to accurately load.",
                UserWarning,
            )
            for key in unserialized_keys:
                unserialized_attr = self.__dict__[key]
                if isinstance(unserialized_attr, BaseSimObj):
                    # Try calling the attr's _to_registry method.
                    # noinspection PyProtectedMember
                    registry, context_dict = unserialized_attr._to_registry(
                        context_dict=context_dict
                    )
                    attribute_dict[key] = registry["id"]
                    continue
                # Try dumping the object using the native JSON
                # serializer.
                try:
                    json.dumps(unserialized_attr)
                except TypeError:
                    warnings.warn(
                        f"Attribute {key} could not be serialized. "
                        f"Dumping the attribute's repr() "
                        f"representation. This attribute will not be "
                        f"fully loaded.",
                        UserWarning,
                    )
                    attribute_dict[key] = [
                        __NOT_SERIALIZED_FLAG__,
                        repr(unserialized_attr),
                    ]
                else:
                    attribute_dict[key] = unserialized_attr

        obj_dict = {"class": obj_type, "attributes": attribute_dict}
        context_dict[obj_id] = obj_dict

        # Check versions of acnportal and certain dependencies.
        # We only need to check the versions if the context dict is
        # empty, indicating the first level of the recursive call.
        acnportal_version, dependency_versions = None, None
        if first_call:
            acnportal_version = pkg_resources.require("acnportal")[0].version
            dependency_versions = {
                "numpy": numpy.__version__,
                "pandas": pandas.__version__,
            }

        return (
            {
                "id": obj_id,
                "context_dict": context_dict,
                "version": acnportal_version,
                "dependency_versions": dependency_versions,
            },
            context_dict,
        )

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Converts the object's attributes into a JSON serializable
        dict. Each ACN-Sim object defines this method differently.

        Args:
            context_dict (Dict[str, JSON Serializable]): Dict mapping
                object ID's to object JSON serializable representations.

        Returns:
            Dict[str, JSON Serializable]: Dict of the object's
                attributes, converted to JSON serializable forms.
            context_dict (Dict[str, JSON Serializable]): Dict mapping
                object ID's to object JSON serializable representations.
                This is returned as context_dict may have been modified
                by this method.
        """
        raise NotImplementedError

    @staticmethod
    def _build_from_id(obj_id, context_dict, loaded_dict=None):
        """
        Given an object ID and a dictionary mapping object ID's to JSON
        serializable representations of ACN-Sim objects, returns the ACN-Sim
        object represented by the given object_id.

        This method is protected (i.e. underscored) so users do not
        directly use this method. PyNoInspection comments have been
        added where this function is used within the package.

        Optionally, a loaded_dict that contains already-loaded objects
        may be provided to avoid duplicated work.

        Args:
            obj_id (str): Object ID of ACN-Sim object to be loaded.
            context_dict (Dict[str, JSON Serializable]): Dict mapping object
                ID's to object JSON serializable representations.
            loaded_dict (Dict[str, BaseSimObj-like]): Dict mapping object
                ID's to ACN-Sim objects.

        Returns:
            BaseSimObj-like: The loaded ACN-Sim object.
            Dict[str, BaseSimObj-like]: Dict mapping object ID's to ACN-Sim
                objects. This is the loaded_dict passed to and modified by
                this function.

        Raises:
            KeyError: Raised if `obj_id` is not found in `context_dict`.
        """
        # Check if this object has already been loaded; return the loaded
        # object if this is the case.
        if obj_id in loaded_dict:
            return loaded_dict[obj_id], loaded_dict

        if obj_id not in context_dict:
            raise KeyError(f"Object with ID {obj_id} not found in context_dict.")

        # Get the class of this object from the context_dict.
        obj_type = context_dict[obj_id]["class"]
        obj_class = locate(obj_type)

        # 'version' is None since we've already checked the version of the
        # parent object.
        # noinspection PyProtectedMember
        obj, loaded_dict = obj_class._from_registry(
            {
                "id": obj_id,
                "context_dict": context_dict,
                "version": None,
                "dependency_versions": None,
            },
            loaded_dict=loaded_dict,
        )

        loaded_dict[obj_id] = obj
        return obj, loaded_dict

    @classmethod
    def from_json(cls, path_or_buf=None):
        """ Returns an ACN-Sim object loaded from in_registry.
        Note URLs have not been tested as path_or_buf input.

        Args:
            path_or_buf (Union[str, FilePathOrBuffer]): a valid JSON
                str, path object or file-like object. Any valid string
                path is acceptable.
        """
        # The code here is from pandas 1.0.1, io.json.from_json(), with
        # modifications.
        filepath_or_buffer, _, _, should_close = get_filepath_or_buffer(path_or_buf)

        exists = False
        if isinstance(filepath_or_buffer, str):
            try:
                exists = os.path.exists(filepath_or_buffer)
            except (TypeError, ValueError):
                pass

        if exists:
            filepath_or_buffer, _ = get_handle(filepath_or_buffer, "r")
            should_close = True

        if isinstance(filepath_or_buffer, str):
            should_close = False
            out_registry = json.loads(filepath_or_buffer)
        else:
            out_registry = json.load(filepath_or_buffer)
        if should_close:
            filepath_or_buffer.close()

        if out_registry["version"] is None:
            warnings.warn(
                f"Missing a recorded version of acnportal in the "
                f"loaded registry. Object may have been dumped with a "
                f"different version of acnportal.",
                UserWarning,
            )
        if out_registry["dependency_versions"] is None:
            warnings.warn(
                f"Missing recorded dependency versions of acnportal in "
                f"the loaded registry. Object may have been dumped "
                f"with different dependency versions of acnportal.",
                UserWarning,
            )

        out_obj = cls._from_registry(out_registry)[0]
        return out_obj

    @classmethod
    def _from_registry(cls, in_registry, loaded_dict=None):
        """
        Returns an object of type `cls` from a JSON serializable
        representation of the object.

        The deserializer is invoked using `cls.from_json()`, but the
        conversion from a serializable representation to an ACN-Sim
        object occurs in this method.

        This method is protected (i.e. underscored) so users do not
        directly use this method. PyNoInspection comments have been
        added where this function is used within the package.

        Deserializer behaviors:

        - Any native ACN-Sim object (that is, any unaltered object
        provided in `acnsim`) that was dumped (serialized) with the
        object's `_to_registry` method loads without error and with
        accurate values for each attribute, assuming the original object
        was of type `cls`.

        - An extension of an ACN-Sim object that appropriately defines a
        `_from_dict` method will load without error and with
        accurate attribute values. This also applies to ACN-Sim
        objects that have as attributes extensions of ACN-Sim objects
        with defined `_from_dict` methods.

        - An extension that doesn't define `_from_dict` may not load
        correctly. The extension will load correctly only if the
        constructor of the object whose `_from_dict` method is called
        takes the same arguments as this object, and any extra
        attributes are JSON-serializable or readable from an ID.

        - A warning is thrown for each attribute not explicitly loaded
        in the deserialization.

        - A warning is thrown for each attribute that is neither naively
        serializable nor readable from an ID or with a `_from_registry`
        method.

        - A warning is thrown if any of acnportal, numpy, or pandas have
        different versions in the current program from the versions
        present when the object was serialized.

        The deserializer returns an object of type `cls`.

        As deserialization of nested objects occurs recursively, this
        method also optionally accepts a `loaded_dict` that contains
        objects (via the ID's) that have already been loaded, to which
        the deserialized object is added.

        Args:
            in_registry (Dict[str, JSON Serializable]): A JSON
                Serializable representation of this object. Takes the
                form:
                    ```
                    {'id': [This object's id],
                     'context_dict': [The dict mapping all object id's
                                      to their JSON serializable
                                      representations]}
                    ```
            loaded_dict (Dict[str, BaseSimObj-like]): Dict mapping
                object ID's to loaded ACN-Sim objects.

        Returns:
            BaseSimObj-like: Loaded object.
            loaded_dict (Dict[str, BaseSimObj-like]): Dict mapping
                object ID's to loaded ACN-Sim objects.

        Raises:
            KeyError: Raised if object represented by `in_json` is not
                found in `context_dict`.

        Warns:
            UserWarning: If the acnportal version of the loaded object
                is different from that of the acnportal doing
                the loading, or if numpy or pandas is a different
                version. If no version is provided ('version' maps
                to None), no warning is raised.
            UserWarning: If any attributes are present in the object
                not handled by the object's `_from_dict` method.
            UserWarning: If any of the unhandled attributes are not
                readable from an ID. This is fine unless this attribute
                is not meant to be JSON serializable, in which case
                the loaded attribute will be incorrect.

        """
        (loaded_dict,) = cls._none_to_empty_dict(loaded_dict)
        obj_id, context_dict, acnportal_version, dependency_versions = (
            in_registry["id"],
            in_registry["context_dict"],
            in_registry["version"],
            in_registry["dependency_versions"],
        )

        # Check current versions of acnportal and certain dependencies
        # against serialized versions.
        if acnportal_version is not None:
            current_version = pkg_resources.require("acnportal")[0].version
            if current_version != acnportal_version:
                warnings.warn(
                    f"Version {acnportal_version} of input acnportal "
                    f"object does not match current version "
                    f"{current_version}."
                )

        if dependency_versions is not None:
            current_dependency_versions = {
                "numpy": numpy.__version__,
                "pandas": pandas.__version__,
            }
            for pkg in dependency_versions.keys():
                if current_dependency_versions[pkg] != dependency_versions[pkg]:
                    warnings.warn(
                        f"Current version of dependency {pkg} does not "
                        f"match serialized version. "
                        f"Current: {current_dependency_versions[pkg]}, "
                        f"Serialized: {dependency_versions[pkg]}.",
                    )

        try:
            obj_dict = context_dict[obj_id]
        except KeyError:
            raise KeyError(f"Object with ID {obj_id} not found in context_dict.")

        # Check if this object has already been converted, and return
        # the appropriate dict if this is the case.
        if obj_id in loaded_dict:
            return loaded_dict[obj_id], loaded_dict

        if obj_dict["class"] != f"{cls.__module__}.{cls.__name__}":
            warnings.warn(
                f"Deserializing as type "
                f"{cls.__module__}.{cls.__name__}. Object was "
                f"serialized as type {obj_dict['class']}.",
                UserWarning,
            )

        # attribute_dict is a dict mapping attribute names to JSON
        # serializable values.
        attribute_dict = obj_dict["attributes"]

        # Call this class' _from_dict method to convert the JSON
        # representation of this object's attributes into the actual
        # object.
        out_obj, loaded_dict = cls._from_dict(attribute_dict, context_dict, loaded_dict)

        # Check that all attributes have been loaded.
        # Warn if some attributes weren't loaded.
        if out_obj.__dict__.keys() != attribute_dict.keys():
            unloaded_attrs = set(attribute_dict.keys()) - set(out_obj.__dict__.keys())
            unrecorded_attrs = set(out_obj.__dict__.keys()) - set(attribute_dict.keys())
            if len(unloaded_attrs) > 0:
                warnings.warn(
                    f"Attributes {unloaded_attrs} present in object of "
                    f"type {obj_dict['class']} but not handled by object's "
                    f"_from_dict method. Loaded object may have inaccurate "
                    f"attributes.",
                    UserWarning,
                )
            if len(unrecorded_attrs) > 0:
                warnings.warn(
                    f"Attributes {unrecorded_attrs} present in object of "
                    f"type {obj_dict['class']} but not recorded in "
                    f"serialization. Loaded object may need to have "
                    f"attributes set. ",
                    UserWarning,
                )
            for attr in unloaded_attrs:
                # Try reading this attribute from an ID in
                # attribute_dict.
                try:
                    out_attr, loaded_dict = cls._build_from_id(
                        attribute_dict[attr], context_dict, loaded_dict=loaded_dict
                    )
                    setattr(out_obj, attr, out_attr)
                except (KeyError, TypeError):
                    if (
                        isinstance(attribute_dict[attr], list)
                        and attribute_dict[attr][0] == __NOT_SERIALIZED_FLAG__
                    ):
                        warnings.warn(
                            f"Loader for attribute {attr} not found. "
                            f"Setting attribute {attr} directly.",
                            UserWarning,
                        )
                        loaded_attr_value = ErrorAllWrapper(attribute_dict[attr][1])
                    else:
                        loaded_attr_value = attribute_dict[attr]
                    # If the attribute was originally JSON serializable,
                    # this is correct loading.
                    try:
                        setattr(out_obj, attr, loaded_attr_value)
                    except AttributeError:
                        # attr could be protected for out_obj. Warn if it is.
                        warnings.warn(
                            f"Attribute {attr} is protected for object of class "
                            f"{out_obj.__class__}. Not setting {attr} to "
                            f"{loaded_attr_value}. Please see {out_obj.__class__} "
                            f"implementation for more info."
                        )

        # Add this object to the dictionary of loaded objects.
        loaded_dict[obj_id] = out_obj
        return out_obj, loaded_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, "BaseSimObj"]] = None,
    ) -> Tuple["BaseSimObj", Dict[str, "BaseSimObj"]]:
        """ Converts a JSON serializable representation of an ACN-Sim
        object into an actual ACN-Sim object.

        Args:
            attribute_dict (Dict[str, JSON Serializable]): A JSON
                Serializable representation of this object's attributes.
            context_dict (Dict[str, JSON Serializable]): Dict mapping
                object ID's to object JSON serializable representations.
            loaded_dict (Dict[str, BaseSimObj-like]): Dict mapping
                object ID's to ACN-Sim objects.

        Returns:
            BaseSimObj-like: Loaded object
            (Dict[str, BaseSimObj-like]): Dict mapping object ID's to
                ACN-Sim objects. This is returned as loaded_dict may
                have been modified by this method.
        """
        raise NotImplementedError
